# daf_sql.py


# import os
# import sys
import re
#import json
#import time
import sqlite3
import functools
    
# no longer need the following due to using pytest
# sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

# from daffodil.lib.daf_types import \  # T_dola, T_dodi, T_la, T_lota, T_doda, T_buff, T_ds, T_lb, T_rli, T_ta, T_lor
                              # # T_ls, T_lola, T_di, T_hllola, T_loda, T_da, T_li, T_dtype_dict, 
                     
import daffodil.lib.daf_utils    as utils
#from daffodil.keyedlist import KeyedList

from typing import List, Dict, Any, Tuple, Optional, Union, cast, Type, Callable, Iterable, Iterator #
def fake_function(a: Optional[List[Dict[str, Tuple[int,Union[Any, str, Callable, Iterable, Iterator, Type]]]]] = None) -> Optional[int]:
    return None or cast(int, 0)       # pragma: no cover

logs = utils                # alias


@functools.lru_cache()
def sql_unesc_str(escaped_name: str) -> str:
    """ Unquote a SQL identifier and unescape embedded double quotes.
    
        Respect any changes to sql_utils.py and daf_sql.py
    """
    escaped_name = escaped_name.strip()
    
    # unencode unusual characters if any are found with double underscores.
    unesc_name = re.sub(r'__([0-9A-Fa-f]{2})', lambda x: chr(int(x.group(1), 16)), escaped_name)
    
    if unesc_name.startswith('"') and unesc_name.endswith('"'):
        return unesc_name[1:-1].replace('""', '"')
        
    return unesc_name

decode_colname_from_sqlcol = sql_unesc_str


# A minimal list of reserved SQL words; you can expand this
_RESERVED_SQL_WORDS_SET = {
    'select', 'from', 'where', 'group', 'order', 'by', 'user', 'value', 'insert',
    'update', 'delete', 'table', 'name', 'join', 'limit'
}

# the following regex identifies identifiers (column names, table names) that are safe and can be unquoted.
_IDENTIFIER_RE = re.compile(r'^[a-z_][a-z0-9_]*$', flags=re.ASCII)


@functools.lru_cache()
def sql_escape_str(name: str, quoting_ok: bool=True) -> str:
    
    """ Escape SQL identifier with minimal quoting:
        - Return as-is if safe
        - Quote if 'quote_ok' is True and it contains uppercase, digits first, symbols, or is a reserved word
            Safely quote a SQL identifier if needed (idempotent and reversible).
        - If not quote_ok, then encode individual characters using double underscore + hex value.
        
        Note: If constructing an INDEX name from a table name will require quoting_ok=False
    
        Respect any changes to sql_utils.py and daf_sql.py
    """

    name = sql_unesc_str(name).strip()  # Normalize quoting and whitespace
    
    if _IDENTIFIER_RE.match(name) and name.lower() not in _RESERVED_SQL_WORDS_SET:
        return name  # Safe: unquoted
        
    if quoting_ok:
        
        new_name = f'"{name.replace(chr(34), chr(34) * 2)}"'  # Quote + escape any embedded quotes
    
    else:
        # THIS VERSION ENCODES INDIVIDUAL CHARS RATHER THAN QUOTING
        """ Encode the original non-compliant characters using double underscores and hex values. 
            this is a reversible encoding. Running this twice will not hurt.
        """
        # 1. escape any illegal characters as __HH where HH is the hex value.
        new_name = re.sub(r'[^0-9A-Za-z_]', lambda x: f'__{ord(x.group()):X}', name)
        
        # 2. escapes the first character of any names with leading numerics.
        # 3. escapes the first character of any reserved words.
        
        if re.search(r'^\d', new_name) or (new_name.lower() in _RESERVED_SQL_WORDS_SET):
            new_name = f"__{ord(new_name[0]):02X}" + new_name[1:]

    return new_name
   

def lod_to_sqlite_table(lod, table_name='tempdata', db_file_path=None, key_col='rowkey'):

    # see also: https://www.sqlite.org/fasterthanfs.html

    if db_file_path is None:
        db_file_path=f'{table_name}.db'

    # Connect to the SQLite database
    conn = sqlite3.connect(db_file_path)
    cursor = conn.cursor()

    # Drop the table if it exists
    cursor.execute(f"DROP TABLE IF EXISTS {table_name}")

    # Extract field names from the first dictionary in the list
    if len(lod) > 0:
        field_names = list(lod[0].keys())
    else:
        raise ValueError("The list of dictionaries is empty.")

    # Create the table using field names as columns
    create_table_query = f"CREATE TABLE IF NOT EXISTS {table_name} ({', '.join(field_names)})"
    cursor.execute(create_table_query)
    
    # Create an index on the key_col column
    if key_col:
        create_index_at_cursor(
                cursor          = cursor,  
                index_colname   = key_col, 
                table_name      = table_name,
                diagnose        = True)


    # Insert data from the list of dictionaries into the table using parameterized queries
    for entry in lod:
        placeholders = ', '.join(['?'] * len(field_names))
        insert_query = f"INSERT INTO {table_name} ({', '.join(field_names)}) VALUES ({placeholders})"
        values = [entry[field] for field in field_names]
        cursor.execute(insert_query, values)

    # Commit changes and close the database connection
    conn.commit()
    conn.close()

# def create_index_at_cursor(cursor, index_colname: str, table_name: str, diagnose: bool=False):

    # # creating the index at the end is more efficient.
    # logs.sts(f"{logs.prog_loc()} creating index for '{index_colname}' on '{table_name}'", 3, enable=diagnose)

    # index_sqlcol = sql_escape_str(index_colname)
    # sql_tablename = sql_escape_str(table_name)
    
    # try:
        # cursor.execute(f"CREATE INDEX IF NOT EXISTS idx_{index_sqlcol} ON {sql_tablename}({index_sqlcol})")
    # except sqlite3.OperationalError as err:
        # if "already exists" in str(err):
            # logs.sts(f"{logs.prog_loc()} index already exists, don't need to add it again.", 3)
        # else:
            # logs.sts(f"Unexpected Error: {err}.", 3)
            # logs.error_beep()
            # breakpoint() #perm
            # pass
    # except Exception as err:        
        # logs.sts(f"Unexpected Error: {err}.", 3)
        # logs.error_beep()
        # breakpoint() #perm
        # pass

    # logs.sts(f"{logs.prog_loc()} Added index of col '{index_colname}' successfully.", 3, enable=diagnose)

def create_index_at_cursor(
        cursor, 
        index_colname:  str, 
        table_name:     str, 
        unique:         bool=False, 
        drop_first:     bool=False, 
        diagnose:       bool=False,
        ) -> bool:                      # success
    """
    Create an index on a specified column in a SQLite table with an optional UNIQUE constraint.
    
    Note: If an index already exists without being unique, this will not replace the index with a unique one.

    Args:
        cursor (sqlite3.Cursor): SQLite cursor to execute the query.
        index_colname (str): The column name to index.
        table_name (str): The table name to create the index on.
        unique (bool): Whether to create a UNIQUE index (default: False).
        drop_first (bool): Whether to drop an existing index first (default: False).
        diagnose (bool): Whether to enable diagnostic logging (default: False).

    Returns:
        bool: True if the index was successfully created or already exists, False otherwise.
        
    Respect any changes both to daf_sql.py and sql_utils.py
    """

    # creating the index at the end is more efficient.
    logs.sts(f"{logs.prog_loc()} creating index for '{index_colname}' on '{table_name}'", 3, enable=diagnose)

    index_sqlcol = sql_escape_str(index_colname)
    quoted_tablename = sql_escape_str(table_name)
    sqlesc_tablename = sql_escape_str(table_name, quoting_ok = False)
    sqlesc_index_col = sql_escape_str(index_colname, quoting_ok = False)
    index_name   = f"idx_{sqlesc_tablename}_{sqlesc_index_col}"
    
    unique_str = 'UNIQUE' if unique else ''
    
    try:
        if drop_first:
            logs.sts(f"{logs.prog_loc()} dropping existing index '{index_name}'", 3, enable=diagnose)
            cursor.execute(f"DROP INDEX IF EXISTS {index_name}")
    
        cursor.execute(f"CREATE {unique_str} INDEX IF NOT EXISTS {index_name} ON {quoted_tablename}({index_sqlcol})")
        cursor.connection.commit()  # Explicit commit after creating the index

        logs.sts(f"{logs.prog_loc()} Added index of col '{index_colname}' successfully.", 3, enable=diagnose)
        return True

    except sqlite3.OperationalError as err:
        if "already exists" in str(err):
            logs.sts(f"{logs.prog_loc()} index already exists, don't need to add it again.", 3)
            return True
        else:
            logs.sts(f"Unexpected OperationalError: {err}.", 3)
    except Exception as err:        
        logs.sts(f"Unexpected Error: {err}.", 3)

    logs.error_beep()
    breakpoint() #perm
    pass
        
    return False


def sum_columns_in_sqlite_table(table_name='tempdata', db_file_path=None):

    if db_file_path is None:
        db_file_path=f'{table_name}.db'
        
    # Connect to the SQLite database
    conn = sqlite3.connect(db_file_path)
    cursor = conn.cursor()

    # Get the column names from the table
    #esc_column_names = sql_utils.get_escaped_columns(cursor, table_name)
    cursor.execute(f"PRAGMA table_info({table_name})")
    column_info = cursor.fetchall()
    esc_column_names = [col[1] for col in column_info]

    # Create a SQL query to calculate the sum for each column
    sum_queries = [f"SUM({col}) AS {col}" for col in esc_column_names]
    query = f"SELECT {', '.join(sum_queries)} FROM {table_name}"

    # Execute the query and fetch the result
    cursor.execute(query)
    result = cursor.fetchone()

    # Close the database connection
    conn.close()

    # Convert the result into a dictionary
    if result:
        # Since column aliases are not supported, we need to manually alias the columns in the result
        sum_dict = dict(zip(esc_column_names, result))
        return sum_dict
    else:
        return None

def get_memory_usage_of_table_in_memory(table_name='tempdata'):
    # Connect to an in-memory SQLite database
    conn = sqlite3.connect(table_name)
    cursor = conn.cursor()

    # Execute a PRAGMA statement to calculate the size of the table
    cursor.execute("PRAGMA page_size;")  # Get page size
    page_size = cursor.fetchone()[0]

    cursor.execute("PRAGMA page_count;")  # Get page count
    page_count = cursor.fetchone()[0]

    # Calculate the total size in bytes
    total_size_bytes = page_size * page_count

    # Close the database connection
    conn.close()

    return total_size_bytes


def print_table_summary(table_name='example', db_file_path=None):

    if db_file_path is None:
        db_file_path=f'{table_name}.db'
        
    # Connect to the SQLite database
    conn = sqlite3.connect(db_file_path)
    cursor = conn.cursor()

    # Query the sqlite_master table to get table information
    cursor.execute(f"SELECT sql FROM sqlite_master WHERE name='{table_name}'")
    table_info = cursor.fetchone()

    # Close the database connection
    conn.close()

    if table_info:
        print(f"Table '{table_name}' summary:")
        print(table_info[0])
    else:
        print(f"Table '{table_name}' not found.")



def sqlite_selectrow(table_name, key_col='rowkey', value='500'): 

    # Connect to the SQLite database
    conn = sqlite3.connect(f"{table_name}.db")
    cursor = conn.cursor()

    # Define the target rowkey value
    target_rowkey = value
    
    # Execute the SQL query to select the row with the specified rowkey
    cursor.execute(f"SELECT * FROM {table_name} WHERE {key_col}=?", (target_rowkey,))
    selected_row = cursor.fetchone()

    # Close the database connection
    conn.close()

    # If a row was found, create a dictionary from the row
    if selected_row:
        column_names = [description[0] for description in cursor.description]
        selected_dict = dict(zip(column_names, selected_row))
        return selected_dict
    else:
        # Return None if no matching row was found
        return None
        




# test_daf_indexing.py

import unittest
from daffodil.daf import Daf, KeyedList  # Assuming your package is named `daffodil`

class TestDaffodilIndexing(unittest.TestCase):
    def setUp(self):
        # Sample data to use in tests
        self.data = [
            ["row1", 2, 3, 4],
            ["row2", 6, 7, 8],
            ["row3", 10, 11, 12],
        ]
        self.cols = ["A", "B", "C", "D"]
        self.keyfield = "A"  # Setting the keyfield to the first column, which contains strings
        self.daf = Daf(lol=self.data, cols=self.cols, keyfield=self.keyfield)

    
    def test_select_cell(self):
        # Test `my_daf["row3", 3]` to select cell at keyfield "row3", col 3
        result = self.daf["row3", 3]
        expected = Daf(lol=[[12]], cols=["D"], keyfield="")  # Keyfield is reset to "" as it is not propagated
        self.assertEqual(result, expected)

    def test_select_cell_value(self):
        # Test `my_daf["row3", 3].to_value()` to get the value at keyfield "row3", col 3
        result = self.daf["row3", 3].to_value()
        expected = 12
        self.assertEqual(result, expected)

    def test_to_dict(self):
        # Test `.to_dict()` method
        result = self.daf["row3"].to_dict()
        expected = {"A": "row3", "B": 10, "C": 11, "D": 12}
        self.assertEqual(result, expected)

    def test_to_list(self):
        # Test `.to_list()` method
        result = self.daf["row3"].to_list()
        expected = ["row3", 10, 11, 12]
        self.assertEqual(result, expected)

    def test_iter_klist(self):
        # Test `.iter_klist()` method
        for i, klist in enumerate(self.daf.iter_klist()):
            expected_keys = ["A", "B", "C", "D"]
            expected_values = self.data[i]

            # Check if it is an instance of KeyedList
            self.assertIsInstance(klist, KeyedList)

            # Verify keys and values
            self.assertEqual(list(klist.keys()), expected_keys)
            self.assertEqual(list(klist.values()), expected_values)

            # Modify the list through the KeyedList
            klist["B"] += 100

        # Ensure changes are reflected in the original data
        self.assertEqual(self.daf["row1", "B"].to_value(), 102)  # First row, column "B" modified
        self.assertEqual(self.daf["row2", "B"].to_value(), 106)  # Second row, column "B" modified

    def test_slice_rows(self):
        # Test slicing rows, e.g., `my_daf[1:3]`
        result = self.daf[1:3]
        expected = Daf(lol=[["row2", 6, 7, 8], ["row3", 10, 11, 12]], cols=self.cols, keyfield=self.keyfield)
        self.assertEqual(result, expected)

    def test_slice_columns(self):
        # Test slicing columns, e.g., `my_daf[:, 1:3]`
        result = self.daf[:, 1:3]
        expected = Daf(lol=[[2, 3], [6, 7], [10, 11]], cols=["B", "C"], keyfield="")
        self.assertEqual(result, expected)

    def test_select_rows_by_key(self):
        # Test selecting specific rows by keyfield values, e.g., `my_daf[["row1", "row3"]]`
        result = self.daf[["row1", "row3"]]
        expected = Daf(lol=[["row1", 2, 3, 4], ["row3", 10, 11, 12]], cols=self.cols, keyfield=self.keyfield)
        self.assertEqual(result, expected)

    def test_select_columns_by_name(self):
        # Test selecting specific columns by names, e.g., `my_daf[:, ["B", "D"]]`
        result = self.daf[:, ["B", "D"]]
        expected = Daf(lol=[[2, 4], [6, 8], [10, 12]], cols=["B", "D"], keyfield="")
        self.assertEqual(result, expected)

    def test_combined_indexing(self):
        # Test combined indexing, e.g., `my_daf[["row1", "row3"], ["B", "D"]]`
        result = self.daf[["row1", "row3"], ["B", "D"]]
        expected = Daf(lol=[[2, 4], [10, 12]], cols=["B", "D"], keyfield="")
        self.assertEqual(result, expected)
        
    def test_row_key_as_tuple(self):
        # Test slicing rows using tuple keys, e.g., `my_daf[("row2", "row3"), :]`
        result = self.daf[("row2", "row3"), :]
        expected = Daf(lol=[["row2", 6, 7, 8], ["row3", 10, 11, 12]], cols=self.cols, keyfield=self.keyfield)
        self.assertEqual(result, expected)

    
    def test_column_key_as_tuple(self):
        # Test slicing columns using tuple keys, e.g., `my_daf[:, ("B", "D")]`
        result = self.daf[:, ("B", "D")]
        expected = Daf(lol=[[2, 4], [6, 8], [10, 12]], cols=["B", "D"], keyfield="")
        self.assertEqual(result, expected)

    def test_keyfield_reset_on_subset(self):
        # Validate keyfield reset when keyfield data is excluded
        result = self.daf[:, ["B", "C"]]
        expected = Daf(lol=[[2, 3], [6, 7], [10, 11]], cols=["B", "C"], keyfield="")
        self.assertEqual(result, expected)

    def test_indexing_with_range(self):
        # Test indexing with range for rows
        result = self.daf[range(1, 3)]
        expected = Daf(lol=[["row2", 6, 7, 8], ["row3", 10, 11, 12]], cols=self.cols, keyfield=self.keyfield)
        self.assertEqual(result, expected)

    def test_full_row_selection(self):
        # Ensure selecting a full row returns the expected Daf
        result = self.daf["row2"]
        expected = Daf(lol=[["row2", 6, 7, 8]], cols=self.cols, keyfield=self.keyfield)
        self.assertEqual(result, expected)
    
    def test_integer_indexing_single_integer(self):
        # Test single integer indexing for rows and columns
        result = self.daf[1, 2]  # Access row 1, column 2
        expected = Daf(lol=[[7]], cols=["C"], keyfield="")
        self.assertEqual(result, expected)

    def test_integer_indexing_range(self):
        # Test integer range indexing for rows and columns
        result = self.daf[1:3, 1:3]  # Access rows 1-2, columns 1-2
        expected = Daf(lol=[[6, 7], [10, 11]], cols=["B", "C"], keyfield="")
        self.assertEqual(result, expected)

    def test_integer_indexing_slice(self):
        # Test slice indexing for rows and columns
        result = self.daf[slice(0, 2), slice(2, 4)]  # Access rows 0-1, columns 2-3
        expected = Daf(lol=[[3, 4], [7, 8]], cols=["C", "D"], keyfield="")
        self.assertEqual(result, expected)

    def test_integer_indexing_list_of_ranges(self):
        # Test list of ranges indexing for rows and columns
        result = self.daf[[range(0, 1), range(2, 3)]]  # Access rows 0,2
        expected = Daf(lol=[
            ["row1", 2, 3, 4],
            ["row3", 10, 11, 12],
            ], cols=["A", "B", "C", "D"], keyfield="A")
        self.assertEqual(result, expected)

    def test_integer_indexing_list_of_integers(self):
        # Test list of integers indexing for rows and columns
        result = self.daf[[0, 2], [1, 3]]  # Access rows 0 and 2, columns 1 and 3
        expected = Daf(lol=[[2, 4], [10, 12]], cols=["B", "D"], keyfield="")
        self.assertEqual(result, expected)
        
    def test_column_indexing_single_integer(self):
        # Test single integer indexing for columns
        result = self.daf[:, 2]  # Access all rows, column 2
        expected = Daf(lol=[[3], [7], [11]], cols=["C"], keyfield="")
        self.assertEqual(result, expected)

    def test_column_indexing_range(self):
        # Test integer range indexing for columns
        result = self.daf[:, 1:3]  # Access all rows, columns 1-2
        expected = Daf(lol=[[2, 3], [6, 7], [10, 11]], cols=["B", "C"], keyfield="")
        self.assertEqual(result, expected)

    def test_column_indexing_slice(self):
        # Test slice indexing for columns
        result = self.daf[:, slice(2, 4)]  # Access all rows, columns 2-3
        expected = Daf(lol=[[3, 4], [7, 8], [11, 12]], cols=["C", "D"], keyfield="")
        self.assertEqual(result, expected)

    def test_column_indexing_list_of_ranges(self):
        # Test list of ranges indexing for columns
        result = self.daf[:, [range(1, 3)]]  # Access all rows, columns 1-2
        expected = Daf(lol=[[2, 3], [6, 7], [10, 11]], cols=["B", "C"], keyfield="")
        self.assertEqual(result, expected)

    def test_column_indexing_list_of_integers(self):
        # Test list of integers indexing for columns
        result = self.daf[:, [0, 2]]  # Access all rows, columns 0 and 2
        expected = Daf(lol=[["row1", 3], ["row2", 7], ["row3", 11]], cols=["A", "C"], keyfield="A")
        self.assertEqual(result, expected)
        
if __name__ == "__main__":
    unittest.main()


# keyedlist.py

from typing import List, Dict, Any, Optional, Union, Iterator, Callable #, Tuple, Type

T_la = List[Any]

import json


class KeyedList:
    """
    KeyedList is a custom data structure in Python that combines the functionality of a dictionary and a list,
    optimized for efficient indexing and manipulation of data. 
    
    It maintains an index for fast key-based access to values stored in a list. 
    This is similar to a conventional dictionary, but the list items are
    not distributed to each item in the dict, but can be an existing list, used without copying.
    
    hd (header dict)
    
    The keys are implemented as a dictionary of indexes, i.e. {'key0': 0, 'key1', 1, ... 'keyn': n} where the keys
    are just examples here. For convenience, we call this structure a "header dict" or 'hd'. 
    To create this the following can be used:
    
        hd = dict(zip(keys, range(len(keys))))
        
    Which is more performant and equivalent to:
    
        hd = {col: idx for idx, col in enumerate(keys)}
        
    But this likely can be improved if a concise standard library function is created, since the current nature
    of a dict is that it is ordered, and thus has an implied index.
    
    Values
    
    For values that are already stored as a list, the list can be adopted as values without copying. An 
    important attribute of this approach is that the parent list is modified if values in the KeyedList 
    are modified, and vice versa. The code should make a copy if the values in the source list need to 
    remain unaltered, or convert to a conventional dict which will inherently make a copy, such as by using .to_dict()
    
    Similarly, if the hd portion already exists, it can be reused on many instances of keyed list. Further, the hd can 
    be used with a list-of-list structure as the column indexes of all rows. If related to such an array, then the
    keys are frequently called 'cols'. See the daffodil package for a full implementation of such a dataframe array.

    Usage:
        - KeyedList can be initialized from either a list of keys and values, a dictionary, or an existing hd and
          list of values, or from another KeyedList.
          
        - It supports standard dictionary operations such as __getitem__, __setitem__, __delitem__, __len__,
          __iter__, keys, values, items, get, update, and conversion to a conventional dictionary using .to_dict().
          
        - KeyedList instances provide fast key-based access to values similar to dictionaries, while also allowing
          list-like operations for efficient value manipulation.
          
        - Most importantly, creation of a KeyedList instance is much faster than creating a conventional dict, because
          the list can be adopted as a reference without copying.

    Initialization:
        - KeyedList(keys_iter, values_list) -> KeyedList: Initialize KeyedList from a list of keys and values_list. 
            Like dict(zip(keys, values_list)) but the values_list is adopted without copying.
            
        - KeyedList(full_dict) -> KeyedList: Initialize KeyedList from a full_dict, a conventional dictionary.
            This method of initialization will copy the values and is relatively expensive.
            
        - KeyedList(keys_iter, default=None) -> KeyedList: Initialize KeyedList from a list of keys and constant 'default' 
            Like dict.fromkeys(keys, default). If unspecified, default is None.
            
        - KeyedList(hd, values_list) -> KeyedList: Initialize KeyedList from an existing hd and values_list.
            This is very fast because no copying occurs. 
    
    Operation
        - A KeyedList instance acts just like a conventional dictionary, but it can be much less expensive to use,
            because the hd portion can be reused, and the values can be adopted without copying. However, there is 
            a slight penalty in access because of the additional indirection.
            
        - If a KeyedList is created from an associated 'record' in a list-of-list (lol) with an associated hd, then
            the KeyedList can be created by reference to the hd and a list in the lol array. Changes to the 
            KeyedList instance will update the lol array, because the list item is actually the same list a the one in the array.
            This behavior is not possible with dictionaries.
            
        - As a result, a KeyedList instance is more 'dangerous' for beginner Python programmers. Dicts are constructed
            always by copying in the keys and values and do not maintain the connection to the prior source of the list
            portion. Instead, a KeyedList instance may may have only references to existing hd and values.
            
        - similar to a dictionary, a KeyedList can provide the keys and values as iterators or lists. The difference is that
            a KeyedList allows assignment to the hd and the list.
            
            klist = KeyedList({'a': 5, 'b' 8})
            print(klist)            # output: {'a': 5, 'b' 8}
            print(klist.keys())     # output: dict_keys(['a', 'b'])
            print(klist.values())   # output: [5, 8]
            alist = klist.values    # grab a reference to the values (no copying)
            alist[1] = 10           # assign a value to the list.
            print(alist)            # output: [5, 10]
            print(klist)            # output: {'a': 5, 'b' 10}  <-- note that klist changes too!
            
            klist._values = [2,3]   # assign new values to the list portion. Assignment like this is NOT supported for dicts.
            print(klist)            # output: {'a': 2, 'b' 3}
            print(alist)            # output: [2, 3]    # note that the list that is a reference to the list portion changes.


        Example2:
            keys = ['a', 'b', 'c']
            values = [1, 2, 3]
            klist = KeyedList(keys, values)     # initialize like you would a dict using dict(zip(keys, values))
            print(klist)                        # Output: {'a': 1, 'b': 2, 'c': 3}
            print(klist['a'])                   # Output: 1
            klist['b'] = 5                      # overwrite a value
            print(klist)                        # Output: {'a': 1, 'b': 5, 'c': 3}
            print(klist.values())               # Output: [1, 2, 3]
            print(klist.values)                 # Output: [1, 2, 3]
           
        See also:
            https://peps.python.org/pep-0412/#alternative-implementation
            
    """

    def __init__(self, 
            arg1: Optional[Union[Dict[Any, Any], List[Any], 'KeyedList']] = None, 
            arg2: Optional[List[Any]] = None,
            default: Optional[Union[int, str, float]] = None,
            ):
            
        if isinstance(arg1, dict):
            if arg2 is None:
                # Case 1: from_dict
                self.hd = type(self)._build_hd(arg1.keys())
                self._values = list(arg1.values())
                return
                
            if isinstance(arg2, list):
                # Case 2: from hd plus values
                if len(arg1) != len(arg2):
                    raise ValueError("keys and values must have the same length")
                self.hd = arg1
                self._values = arg2
                return
            
        elif isinstance(arg1, list) and isinstance(arg2, list):
            # Case 3: from list of keys and values
            if len(arg1) != len(arg2):
                raise ValueError("keys and values must have the same length")
            self.hd = type(self)._build_hd(arg1)
            self._values = arg2
            return
            
        elif isinstance(arg1, list) and arg2 is None:
            # Case 4: from list of keys and default
            self.hd = type(self)._build_hd(arg1)
            self._values = [default] * len(arg1)
            return
            
        elif isinstance(arg1, type(self)) and arg2 is None:
            # Case 5: from keyedlist type
            self.hd = arg1.hd
            self._values = arg1.values
            return
            
        elif arg1 is None and arg2 is None:
            # Case 6, Empty - return a functional empty keyedlist, like {}
            self.hd = {}
            self._values = []
            return
        
        breakpoint() #perm logic error
        pass
        raise ValueError("Must provide either a dict, keys and values, hd and list, or KeyedList")
    
    def __getitem__(self, key):
        """
            indexing can use a scalar key or a list of keys, which returns a list.
        """
        
        if isinstance(key, list):
            return [self._values[self.hd[onekey]] for onekey in key if onekey in self.hd]
            
        elif isinstance(key, (str, int)):
            return self._values[self.hd[key]]

        else:
            raise ValueError
    
    def __setitem__(self, key, value):
        if key not in self.hd:
            # extend hd
            self.hd[key] = len(self.hd)
            
            self._values.append(value)
        else:    
            self._values[self.hd[key]] = value

    
    def __delitem__(self, key):
        index = self.hd.pop(key)
        del self._values[index]
                
        # it is necessary to rebuild hd whenever it is changed.
        self.hd = type(self)._build_hd(self.hd.keys())        
    
    def __len__(self):
        return len(self._values)
    
    def __iter__(self):
        return iter(self.hd)
    
    def keys(self):
        return self.hd.keys()

    @property
    def values(self) -> List[Any]:
        return self._values

    @values.setter
    def values(self, new_values) -> None:
        if isinstance(new_values, list):
            self._values = new_values
            

    def values(self, astype: Optional[Union[Callable, str]]=None) -> List[Any]:
        
        return astype_la(self._values, astype)        
    
    def items(self):
        return ((key, self._values[idx]) for key, idx in self.hd.items())


    def get(self, key, default=None):
        try:
            return self._values[self.hd[key]]
        except Exception:
            return default


    def update(self, other):
        # this could allow direct updating.
        for key, value in other.items():
            self[key] = value
    
    def to_dict(self):
        return dict(self.items())
    
    def __repr__(self):
        return repr(dict(self.items()))
        
    def __bool__(self):
        """ return true if there is something in the values list. """
        return bool(self._values)
        
   
    @staticmethod
    def _build_hd(keys: Iterator):
        # it is necessary to rebuild hd whenever it is changed.
        
        # this is equivalent to:
        
        #   {col: idx for idx, col in enumerate(keys)}
        
        # but this is substantially faster

        return dict(zip(keys, range(len(keys))))
        

    def to_json(self):
        # Serialize KeyedList object to a JSON-compatible dictionary
        return json.dumps({"__KeyedList__": True, "hd": self.hd, "values": self._values})

    @classmethod
    def from_json(cls, json_str) -> 'KeyedList':
        # Deserialize JSON string into a KeyedList object
        obj_dict = json.loads(json_str)
        if "__KeyedList__" in obj_dict and obj_dict["__KeyedList__"]:
            return cls(hd=obj_dict.get("hd", {}), values=obj_dict.get("values", []))
        else:
            raise ValueError("Invalid JSON string for KeyedList")

def astype_la(la: T_la, astype: Optional[Union[Callable, str]]=None) -> T_la:
    """ fix the type according to astype spec if it is not None 
            this function current duplicated in daf_utils
    """        

    if astype is not None:
        if callable(astype): 
            return [astype(val) for val in la]
        elif isinstance(astype, str):
            if astype == 'int':
                return [int(val) for val in la]
            elif astype == 'str':
                return [str(val) for val in la]
            elif astype == 'float':
                return [float(val) for val in la]
            elif astype == 'bool':
                return [bool(val) for val in la]
            else:
                raise ValueError (f"astype not supported: {astype}")
        raise ValueError (f"astype not supported: {astype}")
    return la
        
        
class KeyedListEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, KeyedList):
            return {"__KeyedList__": True, "hd": obj.hd, "values": obj._values}
        return super().default(obj)


    


# daf_indexing.py

# noqa
"""

# Daf -- Python Daffodil dataframes

The Daf class provides a lightweight, simple and fast alternative to provide 
2-d data arrays with mixed types.

This file handles indexing with square brackets[] as functions that operate on
a daf instance.

Indexing is fairly simple in concept, but it has a number of forms and subtle nuances.

THIS FILE IS NO LONGER ACTIVE. INDEXING HAS BEEN SIMPLIFIED AND MOVED TO daf.py, AS IT IS PART OF THE MAIN Daf CLASS.

General Form:   my_daf[row, col]

    1. col parameter is optional. If it is left out, then the row specified includes all columns.
    2. Each of the parameters can be an integer, which is the index into the currently defined lol array.
    3. Each of the parameters can be a slice, i.e. n:m, which denotes a range.
        n is the starting (row or col) index
            if n is omitted, it defaults to 0
        m is 1 past the ending index included in the range.
            if m is omitted, it indicates that all remaining (rows or cols) are included.
        :  without any row or column numbers indicates the entire range.
        
    4. Each of the row, col parameters can be a string.
        each string will select which (row or col) will be included by keyfield or col names.
        my_daf['row5'] -- returns the row with keyfield 'row5' 
        
    5. Each of the row, col parameters can be a list
        A list of integers selects which (rows or cols) will be included.
        A list of strings selects which (rows or cols) will be included by keyfield or col names.
        
    6. If a single cell is selected by row, col by integers or using string selectors, then it is returned as a
        scalar value, such as value = my_daf[4,5] select the cell at location index row 4 and col 5.
        
    8. If both parameters are slices or lists, then a daf array is returned.
        If the columns are not sliced, then the daf returned is not a copy, but a reference to the selected rows in the array.
        
        my_daf[4,5] returns value from that cell in the array, which could be a scalar (default)
        # my_daf[4,5, 'daf'] returns a daf array with one value at 0,0 and associated colname keyfield.
        # my_daf[4,5, 'list'] returns a list with a single value
        # my_daf[4,5, 'dict'] returns a dict with a single item with column name as the key and value.
        
        my_daf[4]   returns a list (default) of row 4, as a reference without making a copy. List is handy for say sum()
        # my_daf[4, :, 'list'] Same as above.
        # my_daf[4, :, 'dict'] returns a dict of row index 4, with colnames as dict keys (cols must be defined)
        # my_daf[4, :, 'daf'] returns a daf array with one row at row index 0, with header and keyfield from my_daf
        
        my_daf[:, 8] returns column at index 8 as a list (default)
        # my_daf[:, 8, 'list'] returns column at index 8 as a list.
        # my_daf[:, 8, 'dict'] return column at index 8 as a dict with keys which are from the keyfield.
        # my_daf[:, 8, 'daf'] return column at index 8 in a new daf at col index 0.
        
        my_daf[:, 'col8'] returns column at colname 'col8' as a list (default)
        # my_daf[:, 'col8', 'list'] returns column at colname 'col8' as a list.
        # my_daf[:, 'col8', 'dict'] returns colname 'col8' as a dict with keys which are from the keyfield.
        # my_daf[:, 'col8', 'daf'] returns a daf with one column with colname 'col8'. This will have no keyfield unless it is included.
        # my_daf[:, ['col0','col8'], 'daf'] returns a daf with two columns, the keyfield column 'col0' and with colname 'col8'. Keyfield is maintained.
        
    # 9. There is a case of possible ambiguity if the keyfield is an integer.
        # It is not clear if it is meant to reference the index of the keyfield. If the row parameter is an 
        # integer, or a list of integers, the default will be to use the integer as an index, and not a key.
        
        # my_daf[4]              returns list at row index 4.
        # my_daf[(4, 'k')]       returns a list of row where keyfield is 4.
        # my_daf[[4,6,8]]        returns a daf of rows at indexes 4,6,8. These references to the rows from the original
        # my_daf[([4,6,8], 'k')] returns a daf of rows where keyfield isin [4,6,8]
        
        # my_daf[:, 4]           returns list of column index 4.
        # my_daf[:, (4, 'k')]    returns a list of column which has colname 4.
        # my_daf[:, (4, 'i')]    returns a list of column at index 4.
        # my_daf[:, ([4,6,8], 'i')] returns a daf of columns at indexes 4,6,8.
        # my_daf[:, ([4,6,8], 'k')] returns a daf of columns where colnames isin [4,6,8]

        # my_daf['asdfj':]
        # my_daf[:'asdfj']
"""

"""
    MIT License

    Copyright (c) 2024 Ray Lutz

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

"""
See README file at this location: https://github.com/raylutz/Daf/blob/main/README.md
"""
# import sys
# from pathlib import Path

# no longer need the following due to using pytest
# sys.path.append(str(Path(__file__).resolve().parents[2]))

from daffodil.lib.daf_types import T_ls, T_li, T_doda, T_lb
# T_lola, T_da, T_di, T_hllola, T_loda, T_dtype_dict, T_dola, T_dodi, T_la, T_lota, T_buff, T_df, T_ds,

import daffodil.lib.daf_utils as utils
# import Daf.daf_md    as md
import daffodil.lib.daf_indexing as indexing  # self import so internal references to indexing will work.
from daffodil.daf import Daf

Daf = Daf   # fool linter

from typing import List, Dict, Any, Tuple, Optional, Union, cast, Type, Callable  #


def fake_function(a: Optional[List[Dict[str, Tuple[int, Union[Any, str, Type, Callable]]]]] = None) -> Optional[int]:
    return None or cast(int, 0)  # pragma: no cover


indexing_version = 'version3'


def _get_item(self,
              slice_spec: Union[slice, int, str, T_li, T_ls, T_lb,
              Tuple[Union[slice, int, str, T_li, T_ls, T_lb],
              Union[slice, int, str, T_li, T_ls, T_lb]]],
              ) -> 'Daf':
    """ allow selection and slicing using one or two specs:
        
            my_daf[2, 3]         -- select cell at row 2, col 3 and return value.
            my_daf[2]            -- select row 2, including all columns, return as a list.
            my_daf[2, :]         -- same as above
            my_daf[[2], :]       -- same as above
            my_daf[[2,5,8]]      -- select rows 2, 5, and 8, including all columns
            my_daf[[2,5,8], :]   -- same as above.
            my_daf[:, 3]         -- select only column 3, including all rows. Return as a list.
            my_daf[:, 'C']       -- select only column named 'C', including all rows, return as a list.
            my_daf[:, ['C']]     -- same as above
            my_daf[:, [1,4,7]]   -- select columns 1,4, and 7, including all rows. Return as a daf
            my_daf[:, ['C','F']] -- select columns 'C' and 'F' including all rows. Return as a daf
            my_daf[[2,5,8], [1,4,7]]     -- select rows 2, 5, and 8, including columns 1,4, and 7.
            my_daf[[2,5,8], ['C','F']]   -- select rows 2, 5, and 8, including columns 'C' and 'F'
            
            my_daf[2:4]          -- select rows 2 and 3, including all columns, return as daf.
            my_daf[2:4, :]       -- same as above
            my_daf[:, 3:5]       -- select columns 3 and 4, including all rows, return as daf.
            
            new row selection using keys:
            my_daf['row1']       -- select entire row with keyfield 'row1'.
                                        Note this differs from Pandas operation.
            my_daf['row1','C']   -- select cell at row with keyfield 'row1' at colname 'C'
                                        Similar to dict-of-dict selection dod['row1']['C']
            my_daf[('row1','row7'), :] -- select rows including 'row1' upto but not including 'row8' (7 rows)
            my_daf[['row1', 'row5, 'row8']]                -- select three rows by list of keyfield names.
            my_daf[('row1', 'row7'), ['C','F']]               -- select rows including 'row1' through 'row7' (7 rows) in columns 'C' and 'F'
            my_daf[['row1', 'row5, 'row8'], ['C','F']]     -- select three rows by list of keyfield names and two columns, by name.
            
        
            returns a consistent daf instance copied from the original, and with the data specified.
            always returns the simplest object possible.
            if multiple rows or columns are specified, they will be returned in the original orientation.
            if only one cell is selected, return a single value.
            If only one row is selected, return a list. If a dict is required, use select_irow()
            if only one col is selected, return a list.
        """

    if indexing_version == 'version2':
        return indexing._get_item2(self, slice_spec)
    elif indexing_version == 'version3':
        return indexing._get_item3(self, slice_spec)

    return indexing._get_item1(self, slice_spec)


def _set_item3(self, slice_spec):
    pass


def _get_item3(self,
               slice_spec: Union[slice, int, str, T_li, T_ls, T_lb,
               Tuple[Union[slice, int, str, T_li, T_ls, T_lb],
               Union[slice, int, str, T_li, T_ls, T_lb]]],
               ) -> 'Daf':
    if isinstance(slice_spec, tuple) and len(slice_spec) == 2:
        # Handle parsing slices for  both rows and columns
        row_spec, col_spec = slice_spec
    else:
        row_spec = slice_spec
        col_spec = None

    if isinstance(row_spec, (int, slice)) or utils.is_list_of_type(row_spec, int):
        sel_rows_daf = self.select_irows(irows=slice_spec)
    elif isinstance(row_spec, str) or utils.is_list_of_type(row_spec, str):
        sel_rows_daf = self.select_krows(krows=slice_spec)
    else:
        return self

    if not col_spec:
        return sel_rows_daf

    if isinstance(col_spec, (int, slice)) or utils.is_list_of_type(col_spec, int):
        return sel_rows_daf.select_icols(icols=slice_spec)
    elif isinstance(col_spec, str) or utils.is_list_of_type(col_spec, str):
        return sel_rows_daf.select_kcols(kcols=slice_spec)
    else:
        return sel_rows_daf


def _get_item1(self,
               slice_spec: Union[slice, int, str, T_li, T_ls, T_lb,
               Tuple[Union[slice, int, str, T_li, T_ls, T_lb],
               Union[slice, int, str, T_li, T_ls, T_lb]]],
               ) -> 'Daf':
    if isinstance(slice_spec, slice):
        # Handle slicing only rows
        return indexing._handle_slice(self, slice_spec, None)

    elif isinstance(slice_spec, int):
        # Handle indexing a single row
        irow = slice_spec
        return indexing._handle_slice(self, slice(irow, irow + 1), None)

    elif isinstance(slice_spec, str):
        # Handle indexing a single row using keyfield
        if not self.keyfield:
            raise RuntimeError("Use of string row spec requires keyfield defined.")
        irow = self.kd.get(slice_spec, -1)
        if irow < 0:
            raise RuntimeError("Row spec not a valid row key.")

        return indexing._handle_slice(self, slice(irow, irow + 1), None)

    elif isinstance(slice_spec, list):
        # single list of row indices
        row_indices = indexing._row_indices_from_rowlist(self, slice_spec)
        return indexing._handle_slice(self, row_indices, None)

    elif isinstance(slice_spec, tuple) and len(slice_spec) == 2:
        # Handle slicing both rows and columns
        return indexing._handle_slice(self, slice_spec[0], slice_spec[1])
    else:
        raise TypeError("Unsupported indexing type. Use slices, integers, or tuples of slices and integers.")


def _handle_slice(self, row_slice: Union[slice, int, None], col_slice: Union[slice, int, str, None]) -> Any:
    """
    Handles the slicing operation for rows and columns in a Daf instance.

    Args:
    - row_slice (Union[slice, int, None]): The slice specification for rows.
    - col_slice (Union[slice, int, None]): The slice specification for columns.

    Returns:
    - Daf: A new Daf instance with sliced data based on the provided row and column specifications.
    or
    a single value if the resulting Daf is only one cell.
    or a list if a single row or col results
    """
    all_cols = self.columns()
    sliced_cols = all_cols
    row_sliced_lol = self.lol  # default is no change

    # handle simple cases first:
    # selecting a row spec is integer.
    if isinstance(row_slice, int):
        irow = row_slice
        # select one cell.
        if isinstance(col_slice, int):
            return self.lol[irow][col_slice]
        elif isinstance(col_slice, str):
            icol = self.hd[col_slice]
            return self.lol[irow][icol]

        # full row    
        if col_slice is None:
            la = self.lol[irow]

        # part of a row, according to a slice
        elif isinstance(col_slice, slice):
            start_col, stop_col, step_col = indexing._parse_slice(self, col_slice, row_or_col='col')
            la = self.lol[irow][start_col:stop_col:step_col]

        # part of a row, according to a column list    
        elif isinstance(col_slice, list):
            # the following handles cases of integer list as well as str list of colnames.
            col_indices_li = indexing._col_indices_from_collist(self, col_slice)
            la = [self.lol[irow][icol] for icol in col_indices_li]
        return la

    # first handle the rows
    elif row_slice is None:
        # no specs, return the entire array.
        if col_slice is None:
            return indexing._adjust_return_val(self)  # is this correct?

        # all rows, one column
        elif isinstance(col_slice, (int, str)):
            if isinstance(col_slice, str):
                icol = self.hd[col_slice]
            else:
                icol = col_slice
            col_la = [row[icol] for row in self.lol]
            return col_la

        # part of all rows, according to a list    
        elif isinstance(col_slice, list):
            col_indices_li = indexing._col_indices_from_collist(self, col_slice)
            row_col_sliced_lol = [[row[i] for i in col_indices_li] for row in self.lol]
            sliced_cols = [all_cols[i] for i in col_indices_li]

        elif isinstance(col_slice, slice):
            # use normal slice approach
            start_col, stop_col, step_col = indexing._parse_slice(self, col_slice, row_or_col='col')
            row_col_sliced_lol = [[row[icol] for icol in range(start_col, stop_col, step_col)] for row in self.lol]
            sliced_cols = [all_cols[icol] for icol in range(start_col, stop_col, step_col)]
            # also respect the column names

        sliced_daf = Daf(lol=row_col_sliced_lol, cols=sliced_cols, dtypes=self.dtypes, keyfield=self.keyfield)
        return sliced_daf._adjust_return_val()

    # sliced or listed rows, first reduce the array by rows.
    elif isinstance(row_slice, list):
        row_indices = indexing._row_indices_from_rowlist(self, row_slice)

        if row_indices:
            row_sliced_lol = [self.lol[i] for i in row_indices]
        # else:
        # row_sliced_lol = self.lol

    elif isinstance(row_slice, slice):
        start_row, stop_row, step_row = indexing._parse_slice(self, row_slice)
        row_sliced_lol = self.lol[start_row:stop_row:step_row]
    # else:
    # row_sliced_lol = self.lol

    # sliced rows, all columns
    if col_slice is None:
        row_col_sliced_lol = row_sliced_lol

    #   one column    
    elif isinstance(col_slice, (int, str)):
        if isinstance(col_slice, str):
            icol = self.hd[col_slice]
        else:
            icol = col_slice

        col_la = [row[icol] for row in row_sliced_lol]
        return col_la

    # part of all sliced rows, according to a list    
    elif isinstance(col_slice, list):
        col_indices_li = indexing._col_indices_from_collist(self, col_slice)
        row_col_sliced_lol = [[row[i] for i in col_indices_li] for row in row_sliced_lol]
        # also respect the column names, if they are defined.
        if self.hd:
            sliced_cols = [all_cols[i] for i in col_indices_li]
        else:
            sliced_cols = None

    elif isinstance(col_slice, slice):

        # use normal slice approach
        start_col, stop_col, step_col = indexing._parse_slice(self, col_slice, row_or_col='col')
        row_col_sliced_lol = [[row[icol] for icol in range(start_col, stop_col, step_col)] for row in row_sliced_lol]
        # also respect the column names, if they are defined.
        if self.hd:
            sliced_cols = [self.columns()[icol] for icol in range(start_col, stop_col, step_col)]
        else:
            sliced_cols = None

    from daffodil.daf import Daf
    sliced_daf = Daf(lol=row_col_sliced_lol, cols=sliced_cols, dtypes=self.dtypes, keyfield=self.keyfield)
    return indexing._adjust_return_val(sliced_daf)


def _adjust_return_val(self):
    # not sure if this is the correct way to go!
    # particularly with zero copy, this may not be correct.
    #
    # alternative is to use .tolist() and .todict()
    # if the method returns a daf.
    #
    # @@TODO: Must be fixed for zero_copy

    num_cols = self.num_cols()
    num_rows = len(self.lol)

    if num_rows == 1 and num_cols == 1:
        # single value, just return it.
        return self.lol[0][0]

    elif num_rows == 1 and num_cols > 1:
        # single row, return as list.
        return self.lol[0]

    elif num_rows > 1 and num_cols == 1:
        # single column result as a list.
        return self.icol(0)

    return self


def _col_indices_from_collist(self, collist) -> T_li:  # col_indices

    if not collist:
        return []

    first_item = collist[0]
    if isinstance(first_item, str):  # probably list of column names (did not check them all)
        colnames = collist
        col_indices = [self.hd[col] for col in colnames]
    elif isinstance(first_item, int):  # probably list of column indices (did not check them all)
        col_indices = collist
    else:
        raise ValueError("column slice, if a list, must be a list of strings or ints")
    return col_indices


def _row_indices_from_rowlist(self, rowlist) -> T_li:  # row_indices
    if not rowlist:
        return []

    first_item = rowlist[0]
    if isinstance(first_item, str):
        if not self.keyfield:
            raise RuntimeError("keyfield must be defined to use str row keys")
        row_indices = [self.kd.get(rowkey, -1) for rowkey in rowlist]
    elif isinstance(first_item, int):
        row_indices = rowlist
    else:
        raise RuntimeError("list spec must use str or int values")

    return row_indices


def _parse_slice(self, s: Union[slice, int, None], row_or_col: str = 'row') -> Tuple[
    Optional[int], Optional[int], Optional[int]]:
    if isinstance(s, slice):
        start = s.start if s.start is not None else 0

        stop = s.stop if s.stop is not None else (self.num_cols() if row_or_col == 'col' else len(self.lol))

        step = s.step if s.step is not None else 1
        return start, stop, step
    elif isinstance(s, int):
        return s, s + 1, 1
    elif s is None:
        return None, None, None


def _set_item(self,
              slice_spec: Union[
                  int, Tuple[Union[slice, int, List[str], List[bool]], Union[slice, int, str, List[str], List[bool]]]],
              value: Any):
    """
        Handles the assignment of values, lists or dicts to Daf elements.
        
        Can handle the following scenarios:
            my_daf[3] = list                   -- assign the entire row at index 3 to the list provided
            my_daf[[3]] = list                 -- same as above
            my_daf[3, :] = list                -- same as above.
            my_daf[[3], :] = list              -- same as above.
            my_daf[3, :] = dict                -- assign the entire row at index 3 to the dict provided, respecting dict keys.
                                                    will only assign those values that have a new value in the provided dict.
            my_daf[3] = value                  -- assign the entire row at index 3 to the value provided.
            my_daf[[3]] = value                -- same as above.
            my_daf[3, :] = value               -- same as above.
            my_daf[[3], :] = value             -- same as above.
            my_daf[[3, 5, 8], :] = value       -- set rows 3, 5 and 8 to the value in all columns
            my_daf[3, 4] = value               -- set one cell 3, 4 to value
            my_daf[3, 5:20] = value            -- set a single value in row 3, columns 5 through 19
            my_daf[3, 5:20] = list             -- set row 3 in columns 5 through 19 to values from the the list provided.
            my_daf[3, [1,4,7]] = list          -- set row 3 in columns 1, 4, and 7 to values from the the list provided.
            my_daf[:, 4] = list                -- assign the entire column at index 4 to the list provided.
            my_daf[3:5, 5] = list              -- assign a partial column at index 5 at rows 3 and 4 to the list provided.

            my_daf[3, 'C'] = value             -- set a value in cell 3, col 'C'
            my_daf[3, ['C', 'D', 'G'] = list   -- set a row 3 in columns 'C', 'D' and 'G' to the values in list.
            my_daf[:, 'C'] = list              -- assign the entire column 'C' to the list provided
            my_daf[3:5, 'C'] = list            -- assign a partial column 'C' to list provided in rows 3 and 4
            my_daf[:, 4] = value               -- assign the entire column 4 to the value provided.
            
            my_daf[[3,5,8], :] = daf          -- set rows 3,5,8 to the data in daf provided
            my_daf[[3,5,8], ['C', 'D', 'G']] = daf   -- set rows 3,5,8 in columns 'C', 'D' and 'G' to the data in daf provided
            my_daf['R1']                       -- choose entire row by name
            my_daf['R1', :]                    -- choose entire row by name
            my_daf[['R1', 'R2', 'R3']]         -- choose three rows
            my_daf[['R1', 'R2', 'R3'], :]      -- choose three rows

        Args:
        - slice_spec: The slice_spec (index or slice) indicating the location to assign the value.
        - value: The value to assign.

        Returns:
        - None
        """

    if indexing_version == 'version2':
        return _set_item2(self, slice_spec, value)
    if indexing_version == 'version3':
        return _set_item3(self, slice_spec, value)
    return _set_item1(self, slice_spec, value)


def _set_item1(self,
               slice_spec: Union[
                   int, Tuple[Union[slice, int, List[str], List[bool]], Union[slice, int, str, List[str], List[bool]]]],
               value: Any):
    if isinstance(slice_spec, int):
        irow = slice_spec
        # Assigning a row based on a single integer index and a value which is a list.
        # will trigger an error if the list is not the right length for the row.
        if isinstance(value, list):
            self.lol[irow] = value

        # if value is a dict, use it and make sure the right columns are assigned per dict keys.    
        elif isinstance(value, dict):
            self.assign_record_irow(irow, record=value)
        else:
            # set the same value in the row for all columns.
            self.lol[irow] = [value] * len(self.lol[irow])

    elif isinstance(slice_spec, list) and isinstance(value, self.__class__):
        # assign a number of rows to the data in daf provided.
        row_indices = self._row_indices_from_rowlist(slice_spec)
        for source_row, irow in enumerate(row_indices):
            self.lol[irow] = value.lol[source_row]

    elif isinstance(slice_spec, tuple):
        row_spec, col_spec = slice_spec
        if isinstance(row_spec, (int, str)):
            if isinstance(row_spec, str) and self.keyfield:
                irow = self.kd[row_spec]
            else:
                irow = row_spec

            if isinstance(col_spec, int):
                # my_daf[irow, icol] = value       -- set cell irow, icol to value, where irow, icol are integers.
                self.lol[irow][col_spec] = value

            elif isinstance(col_spec, str):
                # my_daf[irow, colname] = value    -- set a value in cell irow, col, where colname is a string.
                icol = self.hd[col_spec]
                self.lol[irow][icol] = value

            elif isinstance(col_spec, list) and col_spec:
                col_indices = indexing._col_indices_from_collist(self, col_spec)

                # assign a number of columns specified in a list of colnames to a single row, from a list with only those columns.
                for source_col, icol in enumerate(col_indices):
                    self.lol[irow][icol] = value[source_col]

            elif isinstance(col_spec, slice):
                # my_daf[irow, start:end] = value  -- set a value in cells in row irow, from columns start to end.
                # my_daf[irow, start:end] = list   -- set values from a list in cells in row irow, from columns start to end.
                col_start, col_stop, col_step = indexing._parse_slice(self, col_spec)
                for idx, icol in enumerate(range(col_start, col_stop, col_step)):  # type: ignore
                    if isinstance(value, list):
                        self.lol[irow][icol] = value[idx]
                    else:
                        self.lol[irow][icol] = value

        elif isinstance(row_spec, slice):
            row_start, row_stop, row_step = indexing._parse_slice(self, row_spec)

            if isinstance(col_spec, list) and col_spec:
                col_indices = indexing._col_indices_from_collist(self, col_spec)

                for source_row, irow in enumerate(range(row_start, row_stop, row_step)):
                    for source_col, icol in enumerate(col_indices):
                        self.lol[irow][icol] = value.lol[source_row][source_col]


            elif isinstance(col_spec, (int, str)):
                if isinstance(col_spec, str):
                    icol = self.hd[col_spec]
                else:
                    icol = col_spec

                for idx, irow in enumerate(range(row_start, row_stop, row_step)):  # type: ignore
                    if isinstance(value, list):
                        self.lol[irow][icol] = value[idx]
                    else:
                        self.lol[irow][icol] = value

        elif isinstance(row_spec, str) and self.keyfield:
            irow = self.kd[row_spec]

            if isinstance(col_spec, list) and col_spec:
                col_indices = indexing._col_indices_from_collist(self, col_spec)

                for source_row, irow in enumerate(row_indices):
                    for source_col, icol in enumerate(col_indices):
                        self.lol[irow][icol] = value.lol[source_row][source_col]

            elif isinstance(col_spec, (int, str)):
                if isinstance(col_spec, str):
                    icol = self.hd[col_spec]
                else:
                    icol = col_spec

                for idx, irow in enumerate(range(row_start, row_stop, row_step)):  # type: ignore
                    if isinstance(value, list):
                        self.lol[irow][icol] = value[idx]
                    else:
                        self.lol[irow][icol] = value


        elif isinstance(row_spec, list):
            row_indices = indexing._row_indices_from_rowlist(self, row_spec)

            if isinstance(col_spec, list) and col_spec:
                col_indices = indexing._col_indices_from_collist(self, col_spec)

                for source_row, irow in enumerate(row_indices):
                    for source_col, icol in enumerate(col_indices):
                        self.lol[irow][icol] = value.lol[source_row][source_col]


            elif isinstance(col_spec, (int, str)):
                if isinstance(col_spec, str):
                    icol = self.hd[col_spec]
                else:
                    icol = col_spec

                for idx, irow in enumerate(row_indices):
                    if isinstance(value, list):
                        self.lol[irow][icol] = value[idx]
                    else:
                        self.lol[irow][icol] = value
        else:
            raise ValueError("Unsupported key type for assignment")


    else:
        raise ValueError("Unsupported key type for assignment")

    self._rebuild_kd()


# ===== VERSION 2 ======

def _parse_itemidx(self, slice_spec: Union[slice, int, str, T_li, T_ls, T_lb], row_or_col: str = 'row',
                   parse_doda: Optional[T_doda] = None) -> T_doda:  # parse_doda

    """ parse one index of the item specification using square brackets.
    
        if the return value is a boolean mask, then the rows or columns cannot be reordered.
        It is very costly to reorder columns. Rows can be reordered easily.
        Column order can be altered through a reordering table if desired. i.e. the kd could be
        used to reorder the columns. Therefore, the column calculations should go through 
        the kd to allow column order changes without changing the array.
        
        This function may be called twice in any parsing operation if the type is tuple.
        
    """
    if parse_doda is None:
        parse_doda = {'row': {}, 'col': {}}

    if isinstance(slice_spec, slice):
        if slice_spec.start is None and slice_spec.stop is None:
            # no slicing.
            parse_doda[row_or_col] = {}

        start_idx, stop_idx, step_idx = indexing._parse_slice(self, slice_spec, row_or_col=row_or_col)

        if row_or_col == 'row':
            # save slice tuple directly and do not construct a list of indices
            parse_doda[row_or_col]['slice'] = (start_idx, stop_idx, step_idx)
        else:
            # for col, always resolve to li
            parse_doda[row_or_col]['li'] = [idx for idx in range(start_idx, stop_idx, step_idx)]

    elif isinstance(slice_spec, int):
        if row_or_col == 'row' or not self.kd:
            parse_doda[row_or_col]['li'] = [slice_spec]
        else:
            parse_doda[row_or_col]['li'] = [list(self.hd.values())[slice_spec]]

    elif isinstance(slice_spec, str):
        if row_or_col == 'col':
            icol = self.hd[slice_spec]
            parse_doda[row_or_col]['li'] = [icol]
        else:
            irow = self.kd[slice_spec]
            parse_doda[row_or_col]['li'] = [irow]

    elif isinstance(slice_spec, list) and slice_spec:
        first_item = slice_spec[0]

        if isinstance(first_item, int):
            parse_doda[row_or_col]['li'] = slice_spec

        elif isinstance(first_item, str):
            if row_or_col == 'col':
                colnames_ls = slice_spec
                icol_li = [self.hd.get(colname, -1) for colname in colnames_ls]
                parse_doda[row_or_col]['li'] = icol_li
            else:
                rowkeys_ls = slice_spec
                icol_li = [self.kd.get(rowkey, -1) for rowkey in rowkeys_ls]
                parse_doda[row_or_col]['li'] = icol_li

        elif isinstance(first_item, bool):
            if row_or_col == 'row':
                parse_doda[row_or_col]['lb'] = slice_spec
            else:
                # for col, always resolve to li
                parse_doda[row_or_col]['li'] = [idx for idx, istrue in enumerate(slice_spec) if istrue]

    else:
        raise AttributeError("slice spec error")

    return parse_doda


def _get_parse_doda(self,
                    slice_spec: Union[slice, int, str, T_li, T_ls, T_lb,
                    Tuple[Union[slice, int, str, T_li, T_ls, T_lb],
                    Union[slice, int, str, T_li, T_ls, T_lb]]],
                    ) -> T_doda:  # parse_doda

    if isinstance(slice_spec, tuple) and len(slice_spec) == 2:
        # Handle parsing slices for  both rows and columns
        parse_doda = indexing._parse_itemidx(self, slice_spec[0],
                                             row_or_col='row',
                                             )
        parse_doda = indexing._parse_itemidx(self, slice_spec[1],
                                             row_or_col='col',
                                             parse_doda=parse_doda,
                                             )
    else:
        parse_doda = indexing._parse_itemidx(self, slice_spec,
                                             row_or_col='row',
                                             )
    return parse_doda


def _get_item2(self,
               slice_spec: Union[slice, int, str, T_li, T_ls, T_lb,
               Tuple[Union[slice, int, str, T_li, T_ls, T_lb],
               Union[slice, int, str, T_li, T_ls, T_lb]]],
               ) -> 'Daf':
    parse_doda = indexing._get_parse_doda(self, slice_spec)

    return indexing._get_by_parse_doda(self, parse_doda)


def _set_item2(self,
               slice_spec: Union[slice, int, str, T_li, T_ls, T_lb,
               Tuple[Union[slice, int, str, T_li, T_ls, T_lb],
               Union[slice, int, str, T_li, T_ls, T_lb]]],
               value: Any,
               ) -> 'Daf':
    parse_doda = indexing._get_parse_doda(self, slice_spec)

    return indexing._set_by_parse_doda(self, parse_doda, value)


def _get_by_parse_doda(self, parse_doda: T_doda) -> 'Daf':
    if not parse_doda['row'] and not parse_doda['col']:
        # return all rows and columns unchanged.
        return self

    if self.hd:
        all_cols = sliced_cols = list(self.hd.keys())
    else:
        all_cols = sliced_cols = None

    if parse_doda['row']:
        if 'li' in parse_doda['row']:
            row_indices = parse_doda['row']['li']
            # list of indices specified
            row_sliced_lol = [self.lol[i] for i in row_indices if i >= 0]

        elif 'slice' in parse_doda['row']:
            start, stop, step = parse_doda['row']['slice']
            # list of indices specified
            row_sliced_lol = [self.lol[i] for i in range(start, stop, step)]

        elif 'lb' in parse_doda['row']:
            row_boolean_mask_lb = parse_doda['row']['lb']
            # list of indices specified
            row_sliced_lol = [self.lol[i] for i in range(len(self.lol)) if row_boolean_mask_lb[i]]

    if parse_doda['col']:
        col_indices_li = parse_doda['col']['li']
        row_col_sliced_lol = [[row[i] for i in col_indices_li] for row in row_sliced_lol]
        # also respect the column names, if they are defined.
        if self.hd:
            sliced_cols = [all_cols[i] for i in col_indices_li]

    else:
        row_col_sliced_lol = row_sliced_lol

    new_daf = self.clone_empty(lol=row_col_sliced_lol, cols=sliced_cols)

    return new_daf._adjust_return_val()


def _set_by_parse_doda(self, parse_doda: T_doda, value: Any) -> 'Daf':
    if not parse_doda['row'] and not parse_doda['col']:
        # return all rows and columns unchanged.
        return self

    elif parse_doda['row'] and not parse_doda['col']:
        # no column spec, affect complete rows.

        if 'li' in parse_doda['row'] and isinstance(value, self.__class__):
            row_indices = parse_doda['row']['li']
            # list of indices specified
            for source_row, irow in enumerate(row_indices):
                self.lol[irow] = value.lol[source_row]

        elif 'slice' in parse_doda['row']:
            start, stop, step = parse_doda['row']['slice']
            # slice specified
            for source_row, irow in enumerate(range(start, stop, step)):
                self.lol[irow] = value.lol[source_row]

        elif 'lb' in parse_doda['row']:
            row_boolean_mask_lb = parse_doda['row']['lb']
            # boolean mask specified
            source_row = 0
            for irow, istrue in enumerate(row_boolean_mask_lb):
                if not istrue:
                    continue
                self.lol[irow] = value.lol[source_row]
                source_row += 1
        else:
            raise NotImplementedError

    elif parse_doda['col'] and not parse_doda['row']:
        # affecting only columns.

        dest_col_li = parse_doda['col']['li']
        num_dest_cols = len(dest_col_li)

        if num_dest_cols == 1 and isinstance(value, list):
            icol = dest_col_li[0]
            utils.assign_col_in_lol_at_icol(icol, col_la=value, lol=self.lol)

        # scalar value: repeat in the entire column.  
        elif num_dest_cols == 1 and isinstance(value, (float, int, str)):
            icol = dest_col_li[0]
            utils.assign_col_in_lol_at_icol(icol, col_la=None, lol=self.lol, default=value)

        elif num_dest_cols > 1 and isinstance(value, self.__class__):
            for irow in range(len(value.lol)):
                for source_icol, dest_icol in enumerate(dest_col_li):
                    self.lol[irow][dest_icol] = value.lol[irow][source_icol]
        else:
            raise NotImplementedError

    elif parse_doda['col'] and parse_doda['row']:

        dest_col_li = parse_doda['col']['li']
        num_dest_cols = len(dest_col_li)

        if 'li' in parse_doda['row'] or 'lb' in parse_doda['row']:

            if 'lb' in parse_doda['row']:
                dest_row_li = [idx for idx, istrue in enumerate(parse_doda['row']['lb']) if istrue]
            else:
                dest_row_li = parse_doda['row']['li']

            num_dest_rows = len(dest_row_li)

            if num_dest_rows == 1:
                irow = dest_row_li[0]

                if num_dest_cols == 1:
                    icol = dest_col_li[0]
                    self.lol[irow][icol] = value

                elif num_dest_cols > 1 and isinstance(value, list):
                    for source_icol, dest_icol in enumerate(dest_col_li):
                        self.lol[irow][dest_icol] = value[source_icol]

            elif num_dest_rows > 1 and isinstance(value, self.__class__):
                for source_irow, dest_irow in enumerate(dest_row_li):
                    for source_icol, dest_icol in enumerate(dest_col_li):
                        self.lol[dest_irow][dest_icol] = value.lol[source_irow][source_icol]
            else:
                raise NotImplementedError

        elif 'slice' in parse_doda['row']:
            dest_row_start, dest_row_stop, dest_row_step = parse_doda['row']['slice']
            num_dest_rows = (dest_row_stop - dest_row_start) // dest_row_step

            if num_dest_rows == 1:
                irow = dest_row_start

                if num_dest_cols == 1:
                    icol = dest_col_li[0]
                    self.lol[irow][icol] = value

                elif num_dest_cols > 1 and isinstance(value, list):
                    for source_icol, dest_icol in enumerate(dest_col_li):
                        self.lol[irow][dest_icol] = value[source_icol]

            elif num_dest_rows > 1 and isinstance(value, list):
                for source_irow, dest_irow in enumerate(range(dest_row_start, dest_row_stop, dest_row_step)):
                    for source_icol, dest_icol in enumerate(dest_col_li):
                        self.lol[dest_irow][dest_icol] = value[source_irow]

            elif num_dest_rows > 1 and isinstance(value, self.__class__):
                for source_irow, dest_irow in enumerate(range(dest_row_start, dest_row_stop, dest_row_step)):
                    for source_icol, dest_icol in enumerate(dest_col_li):
                        self.lol[dest_irow][dest_icol] = value.lol[source_irow][source_icol]
            else:
                import pdb
                pdb.set_trace()  # temp
                pass

                raise NotImplementedError

        else:
            raise NotImplementedError

    return self



# daf_pandas.py
"""

# Daf -- Daffodil -- python dataframes.

The Daf class provides a lightweight, simple and fast alternative to provide 
2-d data arrays with mixed types.

This file handles indexing with square brackets[] as functions that operate on
a daf instance 'self'.

"""

"""
    MIT License

    Copyright (c) 2024 Ray Lutz

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""


"""
See README file at this location: https://github.com/raylutz/daffodil/blob/main/README.md
"""
# import os
# import sys
# no longer need the following due to using pytest
#sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__)))))
from daffodil.lib.daf_types import T_df, T_dtype_dict, T_ls #, T_li, T_doda, T_lb
                            # T_lola, T_da, T_di, T_loda, T_dola, T_dodi, T_la, T_lota, T_buff, T_ds, 
                     
import numpy as np
import csv
import io
import pandas as pd
# import daffodil.lib.daf_utils    as daf_utils


from typing import List, Dict, Any, Tuple, Optional, Union, cast, Type, Callable #
def fake_function(a: Optional[List[Dict[str, Tuple[int,Union[Any, str, Type, Callable ]]]]] = None) -> Optional[int]:
    return None or cast(int, 0)   # pragma: no cover

# define a sentinel object to express a missing item where None is a valid value.
from .daf_utils import _MISSING

#==== Pandas
@classmethod
def _from_pandas_df(
        cls,
        df: T_df, 
        keyfield: str='', 
        name: str='', 
        use_csv: bool=False, 
        dtypes: Optional[T_dtype_dict]=None
        ):
    """
    Convert a Pandas dataframe to daf object
        @@TODO: This does not enforce dtypes are correct.
    """
    import pandas as pd     # type: ignore
    
    python_dtypes = pandas_dtype_to_python(df)

    if isinstance(df, pd.Series) or not use_csv:

        if isinstance(df, pd.Series):
            rowdict = df.to_dict()
            cols = list(rowdict.keys())
            lol = [list(rowdict.values())]
        else:
            cols = list(df.columns)
            lol = df.values.tolist()

        return cls(cols=cols, lol=lol, keyfield=keyfield, name=name, dtypes=python_dtypes)
        
    # first convert the Pandas df to a csv buffer.
    try:
        csv_buff = df.to_csv(None, index=False, quoting=csv.QUOTE_MINIMAL, lineterminator= '\r\n')
    except TypeError:   # pragma: no cover
        # this uses the old version of the lineterminator with an underscore. 
        csv_buff = df.to_csv(None, index=False, quoting=csv.QUOTE_MINIMAL, line_terminator= '\r\n')
        
    return cls.from_csv_buff(
        csv_buff=csv_buff,
        keyfield=keyfield,
        dtypes=python_dtypes,    
        unflatten=False,  
        )
        

def _to_pandas_df(
    self, 
    cols: Optional[T_ls] = None,
    *,
    use_csv: bool = False, 
    use_donpa: bool = False, 
    default: Any = _MISSING,
    defaulting_cols: Optional[T_ls] = None,
    ) -> Any:
    """
    Convert the Daffodil table to a Pandas DataFrame.

    Parameters:
        cols:
            List of columns (names or indices) to include in the DataFrame.
            If None, includes all columns.

        default:
            If provided, replaces all '' and None with this value.
            Requires that defaulting_cols be provided to indicate which columns to clean.

        defaulting_cols:
            Columns to which default should apply. Ignored if default is not provided.
            If None and default is given, default applies to all included columns.

        use_csv:
            Convert via internal CSV buffer. Faster in some cases.

        use_donpa:
            Convert via dict-of-numpy-arrays (Daffodil-native, fastest for numeric).
            Probably not a good choice for arrays with non-numeric columns.

    Returns:
        Pandas DataFrame.


    Notes:
    - Daffodil uses '' (empty string) as the standard representation for unset or missing values,
      which is ideal for display and printing. These will appear as strings in the resulting DataFrame.
    - Pandas treats '' as a valid string, not a missing value, which can cause columns to be inferred
      as 'object' type even if most values are numeric.
    - This method avoids implicit type coercion: string values like '123' remain strings,
      even if they look numeric. It is up to the caller to post-process columns if explicit type
      conversion is desired.
    """

    selected_cols = cols if cols is not None else self.columns()
    default_cols = defaulting_cols if defaulting_cols is not None else selected_cols

    if use_donpa:
        donpa = self.to_donpa(selected_cols, default=default)
        df = pd.DataFrame(donpa)
    elif use_csv:
        csv_buff = self.to_csv_buff()
        sio = io.StringIO(csv_buff)
        df = pd.read_csv(sio, na_filter=False, index_col=False)
        
        # No Daffodil-native defaulting possible in this path
        if default is not _MISSING:
            raise NotImplementedError("default + use_csv=True is not supported with Daffodil-native logic.")
            
    else:
        # Apply default substitution before DataFrame conversion
        if default is not _MISSING:
            self.replace_in_columns(default_cols, ['', None], default)
       
        df = pd.DataFrame(self[:, selected_cols], columns=selected_cols)

    return df
        

def pandas_dtype_to_python(df: T_df) -> Optional[Any]:
    """
    Translate a Pandas data type to its equivalent Python data type.

    Args:
        pandas_dtype (Any): The Pandas data type to translate.

    Returns:
        Optional[Type]: The equivalent Python data type.
    """


    if isinstance(df, pd.Series):
        # For Series, it is necessary to examine each element
        series_dtypes = dtypes_dict_from_series(df)
        python_dtypes = pandas_dtype_dict_to_python(series_dtypes)
        
    else:
        python_dtypes = pandas_dtype_dict_to_python(df.dtypes.to_dict())
        
    return python_dtypes


def pandas_dtype_dict_to_python(pandas_dtype_dict: Any) -> Optional[Any]:
    """
    Translate a Pandas data type to its equivalent Python data type.

    Args:
        pandas_dtype (Any): The Pandas data type to translate.

    Returns:
        Optional[Type]: The equivalent Python data type.
    """
    python_dtype_dict = {}
    for colname, pandas_dtype in pandas_dtype_dict.items():
        if pandas_dtype == np.object_ or pandas_dtype is str:
            python_dtype_dict[colname] = str
        elif pandas_dtype in (np.int64, np.int32):
            python_dtype_dict[colname] = int
        elif pandas_dtype == np.float64:
            python_dtype_dict[colname] = float
        elif pandas_dtype == np.bool_:
            python_dtype_dict[colname] = bool
        elif pandas_dtype == np.datetime64:
            python_dtype_dict[colname] = pd.Timestamp
        elif pandas_dtype == np.timedelta64:
            python_dtype_dict[colname] = pd.Timedelta
        else:
            print(f"Unknown Pandas dtype for column '{colname}': {pandas_dtype}")
            breakpoint() #perm
            pass
            
    return python_dtype_dict


def python_dtype_to_pandas(python_type: Type) -> Optional[Any]:
    """
    Translate a Python data type to its equivalent Pandas data type.

    Args:
        python_type (Type): The Python data type to translate.

    Returns:
        Optional[Any]: The equivalent Pandas data type.
    """
    dtype_mapping = {
        str: "object",
        int: "int64",
        float: "float64",
        bool: "bool",
        pd.Timestamp: "datetime64[ns]",
        pd.Timedelta: "timedelta64[ns]",
        pd.Categorical: "category",
    }
 
    return dtype_mapping.get(python_type, None)
    

def dtypes_dict_from_series(series: pd.Series) -> Dict[str, type]:
    """
    Construct a dictionary mapping column names to their respective data types from a Pandas Series.

    Args:
        series (pd.Series): The Pandas Series.

    Returns:
        Dict[str, type]: A dictionary mapping column names to their respective data types.
    """
    
    dtypes_dict = {}
    for column_name in series.index:
        dtype = series.dtype
        dtype_name = np.dtype(dtype).name  # Get the name of the NumPy dtype
        dtype_type = type(series[column_name])
        dtypes_dict[column_name] = dtype_type if dtype_name == 'object' else dtype_type.__base__
    return dtypes_dict
    


# daf_md.py -- markdown generation

# copyright (c) 2025 Ray Lutz

import re
#import copy
import markdown
#import functools

#import pprint
from string import Template

# import os
# import sys
# no longer need the following due to using pytest
# sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__)))))
from daffodil.lib.daf_types import T_ls, T_lola, T_da, T_li #, T_doda, T_df, T_lf, T_loda, T_loloda, T_lodolodi, T_ts, T_ds, T_dola 
import daffodil.lib.daf_utils as utils

from typing import List, Dict, Any, Tuple, Optional, Union #, cast
def fake_function(a: Optional[List[Dict[str, Union[int, Tuple[int, Any]]]]] = None) -> Optional[int]:
    return None   # pragma: no cover


# Translation dictionaries for table alignment
left_rule = {'<': ':', '^': ':', '>': '-'}
right_rule = {'<': '-', '^': ':', '>': ':'}

#====================================================================================
#    Markdown Content Retrieval Function

# @functools.lru_cache(maxsize=None)  # Apply caching to load_locale_data
# def load_locale_ds(locale: str) -> T_ds:
    # """ load localized text strings.
        # Use remove_newlines when putting text in the cell of a table, for example.
    # """

    # # Load the appropriate @section@ marked Markdown file based on the locale
    # markdown_content = DB.load_data(dirname='localized_text', name=f"mdtext_{locale}.md", s3flag=False)

    # # Define a regular expression pattern to match section headers
    # pattern = r'^%SECTION_NAME%\s*=\s*"([^"]+)"\s*$'

    # # Use re.split to split the content based on section headers
    # # this provides both the headers and the content between the headers.
    # sections_ls = re.split(pattern, markdown_content, flags=re.M)
    
    # # remove any leading & trailing newlines around the block.
    # locale_ds = {sections_ls[idx]: sections_ls[idx+1] for idx in range(1, len(sections_ls), 2)}
    
    # return locale_ds
    

# def getmd(section_name:str, locale:str='', default:str='', remove_newlines: bool=False):
    # """
    # Markdown Content Retrieval Function

    # This function, getmd, allows you to retrieve Markdown content based on the section name 
    # and the selected locale. It loads the appropriate Markdown file, parses it using an 
    # INI-style structure, and caches the results to avoid unnecessary file access when 
    # called multiple times.

    # Usage:
        # content = md.getmd(section_name, locale)

    # Parameters:
        # section_name (str): The name of the section to retrieve.
        # locale (str): The selected locale for content retrieval.
        # default (str): used if no record is found.

    # Example:
        # locale = 'en_US'
        # section_name = 'Section 1'
        # content = getmd(section_name, locale)
        # print(content)
 
    # # Example usage:
    # content = getmd(section_name='Section 1', locale = 'en_US')
    # print(content)
    
    # Variables can be included in the text in {} curly braces.
    # To substitute, use this syntax:
    # content = md.getmd('section name').format(argbane=argvalue, ...)
 
    # """

    # # commented out for Daf.
    # # if not locale:
        # # locale = args.argsdict.get('locale', 'en_US') or 'en_US'

    # locale_ds = load_locale_ds(locale)

    # # Retrieve the section content from the locale data
    # if section_name not in locale_ds:
        # print(f"WARN: Can't find section '{section_name}' in locale {locale} text data.")

        # if locale != 'en_US':
            # locale_ds = load_locale_ds('en_US')
            # section_content = locale_ds.get(section_name, default)
        # else:    
            # section_content = default
    # else:
        # section_content = locale_ds[section_name]
        
    # if remove_newlines:
        # section_content = re.sub(r'[\n\r]+', ' ', section_content, flags=re.S)
        # # section_content = re.sub(r'<br>',    ' ', section_content).strip()
        # section_content = re.sub(r'\s\s+',   ' ', section_content)

    # return section_content
    

def mdlink(url_or_s3path: str='', title: str='', new_window: bool=False) -> str:
    if not url_or_s3path:
        return f"{title} (Not Available)"
        
    if url_or_s3path.startswith('s3'):    
        if not title:
            title = utils.safe_basename(url_or_s3path)
            url   = utils.s3path_to_url(url_or_s3path)
    else:
        url = url_or_s3path
        
    if new_window:
        return new_window_link(url, title)
    return f"[{title}]({url})"


def mdlink_s3path(s3path: str='', title: str='', new_window: bool=False) -> str:

    return mdlink(url_or_s3path=s3path, title=title, new_window=new_window)


def new_window_link(url: str='', title: str=''):
    return f'<a href="{url}" target="_blank">{title}</a>'
    

def escape_internal_link(header_text: str) -> str:
    # remove all special characters.
    # convert spaces to '-'
    #
    # to create a link: f"['local text'](#{md.escape_internal_link('header text to link to')})
    #
    # <h2 id="precinct-v-windsor-wds-1-2">Precinct V Windsor Wds 1-2</h2>

    cleaned_text = re.sub(r'[^\w_\s\-]', '', header_text)   # remove all non-word, underscores, spaces and hyphens.
    internal_link = cleaned_text.lower().strip().replace(' ', '-')
    internal_link = re.sub(r'--', '-', internal_link)
    internal_link = re.sub(r'--', '-', internal_link)   # just in case there were overlapping patterns missed above (it happens)
    
    return internal_link
    
    
def md_parse_link(md_link: str) -> Tuple[str, str]:

    # Usage:
    #   text, link = md_parse_link(md_link)

    text = md_link
    link = ''
    if md_link.startswith('['):
        match = re.search(r'^\[([^\]]+)\]\(([^)]+)\)$', md_link)    
        text = match[1]             # type: ignore
        link = match[2]             # type: ignore
        
    return text, link
        

def md_toc(headings_list: T_ls):

    rep = "# Table of Contents\n\n"
    
    for heading in headings_list:
        rep += f"- [{heading}](#{escape_internal_link(heading)})\n"  
    return rep + "\n\n"

# called by Daf.to_md()

def md_lol_table(
        records_lol:        T_lola, 
        header:             Optional[T_ls]=None, 
        includes_header:    bool=False, 
        align:              Optional[List[Tuple[str, str]]]=None, 
        just:               str='', 
        omit_header:        bool=False, 
        shorten_text:       bool=True, 
        max_text_len:       int=80, 
        smart_fmt:          bool=False,
        include_idx:        bool=False,
        sum_col_idxs:       Optional[T_li]=None,
        ):
    """
    Generate a Doxygen-flavor Markdown table from records.
    This could be called md_lol_table()

    records -- list of list of strings
    header -- True if first line should be the heading
    alignment - List of pairs alignment characters.  The first of the pair
        specifies the alignment of the header, (Doxygen won't respect this, but
        it might look good, the second specifies the alignment of the cells in
        the column.

        Possible alignment characters are:
            '<' = Left align (default for cells)
            '>' = Right align
            '^' = Center (default for column headings)
    """
    if not records_lol:
        return ''

    # reorient data into columns
    # would like to not shorten if shorten_text=False but not sure how to do it here.
    # the reason is that the column is set to the length of the entire string rather than longest line
    
    # first remove the header if there is one.
    if includes_header:
        header = records_lol[0]
        records_lol=records_lol[1:]

    if not records_lol:
        return ''    

    if sum_col_idxs:
        new_row = utils.lol_sum_cols(records_lol, sum_col_idxs=sum_col_idxs)
        records_lol += [new_row]

    try:
        cols_lol = utils.transpose_lol(records_lol)
    except RuntimeError as err:
        print(f"{err}: records:{records_lol}")
        #logs.error_beep()
        breakpoint() #perm -- debugging aide
        pass
        
    if include_idx:
        # there is never a header at this point.
        col_idx_ls = [str(idx) for idx in range(len(records_lol))]                        
        cols_lol = [col_idx_ls] + cols_lol
        if header:
            header = ['idx'] + header
        just = '^' + just

    return md_cols_lol_table(cols_lol, header=header, align=align, just=just, 
                omit_header=omit_header, shorten_text=shorten_text, max_text_len=max_text_len, smart_fmt=smart_fmt)


def md_cols_lol_table(
        cols_lol:       T_lola, 
        header:         Optional[T_ls] = None, 
        align:          Optional[List[Tuple[str, str]]] = None, 
        just:           str='', 
        omit_header:    bool = False, 
        shorten_text:   bool = True,
        max_text_len:   Optional[int] = None,
        smart_fmt:      bool = False,
        ):

    """ Use this function when a number of columns of data already exist that should be listed side by side.
        This function does not include the header in the data because it is easier to provide a separate list of strings.
    """

    if not cols_lol:
        return ''
    if max_text_len is None:
        max_text_len = 80
        
    #mutated_cols_lol = copy.deepcopy(cols_lol)
    mutated_cols_lol = cols_lol

    # either get the header or pull it off the columns
    if omit_header:
        pass
    elif header:
        # add header to the columns
        header = [f"{col}".replace('\n', '<br>').replace('\r', '') for col in header]
        for idx, col in enumerate(mutated_cols_lol):
            s = utils.safe_get_idx(header, idx, default='')
            col.insert(0, s)
    else:
        raise RuntimeError ("md header is required if 'omit_header' not specified.")
    
    num_cols = len(mutated_cols_lol)
    max_row = max([len(col) for col in mutated_cols_lol], default=0)
    
    # fill new_cols with blanks to start with.
    new_cols_lol = []
    for _ in range(num_cols):
        new_cols_lol.append([''] * max_row)

    # sanitize cols
    for col_idx, col in enumerate(mutated_cols_lol):
        pass
        for row_idx in range(max_row):
            try:
                val = col[row_idx]
            except Exception:
                val = ''
            if smart_fmt:
                val = utils.smart_fmt(val)
            else:    
                val = f"{val}"
            val = val.replace('\n', '<br>').replace('\r', '')
            
            if shorten_text:
                val = utils.shorten_str_keeping_ends(val, max_text_len)
            new_cols_lol[col_idx][row_idx] = escape_md_table_text(val)

    # cols_lol = new_cols_lol

    # Fill out any missing alignment characters.
    just_tup: Tuple[str, ...] = ()
    if not just and not align:
        just = '^' * num_cols
    if just and not align:
        just_tup = tuple(just)
    if just_tup and not align:    
        align = []
        for just_char in just_tup:
            align.append(('^', just_char))

    extended_align = align if align is not None else []
    extended_align = extended_align[0:num_cols]
    if len(extended_align) < num_cols:
        extended_align += [('^', '<')
                           for i in range(num_cols-len(extended_align)) ]

    heading_align, cell_align = [x for x in zip(*extended_align)]

    # here field width must take into account if the lines have breaks, and if so, then calculate for longest line.
    field_widths = []
    for column in new_cols_lol:
        w_max = 0
        for text in column:
            lines = text.split('<br>')
            for line in lines:
                w = len(line)
                if w > w_max:
                    w_max = w
        field_widths.append(w_max)
    
    # now that the columns have been formatted, remove the header from the columns.
    
    heading_template    = '| ' + ' | '.join([f"{{:{a}{w}}}" for a, w in zip(heading_align, field_widths)]) + ' |'
    row_template        = '| ' + ' | '.join([f"{{:{a}{w}}}" for a, w in zip(cell_align, field_widths)])    + ' |'
    ruling              = '| ' + ' | '.join([left_rule[a] + '-'*(w-2) + right_rule[a] for a, w in zip(cell_align, field_widths)]) + ' |'

    ls = []
    if not omit_header:
        heading = [col.pop(0) for col in new_cols_lol]
        if heading:
            ls.append( heading_template.format(*heading).rstrip() + '\n' )
        ls.append( ruling.rstrip() + '\n' )
    for row in zip(*new_cols_lol):
        ls.append( row_template.format(*row).rstrip() + '\n')
    return ''.join(ls)
        
       
    
def escape_md_table_text(s: str) -> str:
    """ this function escapes text inside a table minimally to allow the markdown table to be built.
    """
    s = s.replace('|', r'\|').replace("\n", '<br>')
    return s
    
    
def escape_raw_text(text: str='') -> str:
    """ Escape raw text so markdown does not recognise any special characters
        Good for regex strings to allow any characters in a raw string.
    """
    
    # summary:
    # regex: 12.338
    # build from list: 4.442
    # build from set: 3.38
    # build from string: 2.225
    
    
    # for example:
    # md.escape_raw_text(r'^(.*)\s*\(.+\)\s*$')
    # r'^\(\.\*\)\\s\*\\\(\.\+\\\)\\s\*$'
    
    if not text:
        return ''
        
    # # character build using list method:
    # # timeit.timeit("md.escape_raw_text(r'^(.*)\s*\(.+\)\s*$')", setup="from utilities import md")
    # # 4.442 secs for 1,000,000 iterations
            
    # spec_char_list = ['\\', '`', '*', '_', '{', '}', '[', ']', '(', ')', '#', '+', '-', '.', '!', '/', '|']
    # output_str = ''
    # for ch in text:
        # if ch in spec_char_list:
            # output_str += '\\' + ch
        # else:
            # output_str += ch

    # character build using set
    # 3.38 secs
    # spec_char_set = {'\\', '`', '*', '_', '{', '}', '[', ']', '(', ')', '#', '+', '-', '.', '!', '/', '|'}
    # output_str = ''
    # for ch in text:
        # if ch in spec_char_set:
            # output_str += '\\' + ch
        # else:
            # output_str += ch


    # character build using string method
    # 2.225 secs for 1,000,000 iterations
    spec_chars = r'\`*_{}[]()#+-.!/|'
    output_str = ''
    for ch in text:
        if ch in spec_chars:
            output_str += '\\' + ch
        else:
            output_str += ch


    
            
    # regex method
    # 12.338 secs for 1,000,000 iterations
    
    # output_str = re.sub(r'([\\`*_{}\[\]()\#+\-\.!/|])', r'\\\1', text)

    return output_str
    
    
def md_process_template(template_str: str, mapping: T_da):

    # replace occurrences of ${dict_key} with values from mapping (dict)
    
    t = Template(template_str)
    identifiers = t.get_identifiers()   # type: ignore # Returns a list of the valid identifiers in the template, 
                                        # in the order they first appear, ignoring any invalid identifiers.
    mapping_keys_set = set(mapping.keys())
    identifiers_set  = set(identifiers)
    
    if mapping_keys_set != identifiers_set:
        print(f"There is a difference between the identifiers in the template:{identifiers_set} vs. the map:{mapping_keys_set}")
        print(f"The template appears to be {'valid' if t.is_valid() else 'invalid'}.")    # type: ignore
    
    return t.safe_substitute(mapping=mapping)


def md_2_html_snippet(md: str, strip_newlines: bool=True):
    #print("Parsing md to html:\n"+md+"\n")
    snippet = markdown.markdown(md, extensions=['tables','toc'])
    if strip_newlines:
        snippet = snippet.replace('\n', '')
        
    return snippet
    
    
def md_2_html(title: str, md: str, strip_newlines: bool=False, include_open_image_js: bool=False):

    md_html = md_2_html_snippet(md, strip_newlines=strip_newlines)


    return f"""
<!DOCTYPE html>
<html>
<head>
<meta http-equiv="Content-Type" content="text/html; charset=utf-8"/>

<title>{title}</title>

<style type="text/css">
{stylesheet}
</style>
{open_image_js if include_open_image_js else ''}
{identical_values_js}
</head>

<body>{md_html}</body></html>"""

stylesheet = """
body {
  font-family: Helvetica, arial, sans-serif;
  font-size: 14px;
  line-height: 1.6;
  padding-top: 10px;
  padding-bottom: 10px;
  background-color: white;
  padding: 30px;
  color: #333;
}

body > *:first-child {
  margin-top: 0 !important;
}

body > *:last-child {
  margin-bottom: 0 !important;
}

a {
  color: #4183C4;
  text-decoration: none;
}

a.absent {
  color: #cc0000;
}

a.anchor {
  display: block;
  padding-left: 30px;
  margin-left: -30px;
  cursor: pointer;
  position: absolute;
  top: 0;
  left: 0;
  bottom: 0;
}

h1, h2, h3, h4, h5, h6 {
  margin: 20px 0 10px;
  padding: 0;
  font-weight: bold;
  -webkit-font-smoothing: antialiased;
  cursor: text;
  position: relative;
}

h2:first-child, h1:first-child, h1:first-child + h2, h3:first-child, h4:first-child, h5:first-child, h6:first-child {
  margin-top: 0;
  padding-top: 0;
}

h1:hover a.anchor, h2:hover a.anchor, h3:hover a.anchor, h4:hover a.anchor, h5:hover a.anchor, h6:hover a.anchor {
  text-decoration: none;
}

h1 tt, h1 code {
  font-size: inherit;
}

h2 tt, h2 code {
  font-size: inherit;
}

h3 tt, h3 code {
  font-size: inherit;
}

h4 tt, h4 code {
  font-size: inherit;
}

h5 tt, h5 code {
  font-size: inherit;
}

h6 tt, h6 code {
  font-size: inherit;
}

h1 {
  font-size: 28px;
  color: black;
}

h2 {
  font-size: 24px;
  border-bottom: 1px solid #cccccc;
  color: black;
}

h3 {
  font-size: 18px;
}

h4 {
  font-size: 16px;
}

h5 {
  font-size: 14px;
}

h6 {
  color: #777777;
  font-size: 14px;
}

p, blockquote, ul, ol, dl, li, table, pre {
  margin: 15px 0;
}

hr {
  border: 0 none;
  color: #cccccc;
  height: 4px;
  padding: 0;
}

body > h2:first-child {
  margin-top: 0;
  padding-top: 0;
}

body > h1:first-child {
  margin-top: 0;
  padding-top: 0;
}

body > h1:first-child + h2 {
  margin-top: 0;
  padding-top: 0;
}

body > h3:first-child, body > h4:first-child, body > h5:first-child, body > h6:first-child {
  margin-top: 0;
  padding-top: 0;
}

a:first-child h1, a:first-child h2, a:first-child h3, a:first-child h4, a:first-child h5, a:first-child h6 {
  margin-top: 0;
  padding-top: 0;
}

h1 p, h2 p, h3 p, h4 p, h5 p, h6 p {
  margin-top: 0;
}

li p.first {
  display: inline-block;
}

ul, ol {
  padding-left: 30px;
}

ul :first-child, ol :first-child {
  margin-top: 0;
}

ul :last-child, ol :last-child {
  margin-bottom: 0;
}

dl {
  padding: 0;
}

dl dt {
  font-size: 14px;
  font-weight: bold;
  font-style: italic;
  padding: 0;
  margin: 15px 0 5px;
}

dl dt:first-child {
  padding: 0;
}

dl dt > :first-child {
  margin-top: 0;
}

dl dt > :last-child {
  margin-bottom: 0;
}

dl dd {
  margin: 0 0 15px;
  padding: 0 15px;
}

dl dd > :first-child {
  margin-top: 0;
}

dl dd > :last-child {
  margin-bottom: 0;
}

blockquote {
  border-left: 4px solid #dddddd;
  padding: 0 15px;
  color: #777777;
}

blockquote > :first-child {
  margin-top: 0;
}

blockquote > :last-child {
  margin-bottom: 0;
}

table {
  padding: 0;
}
table tr {
  border-top: 1px solid #cccccc;
  background-color: white;
  margin: 0;
  padding: 0;
}

table tr:nth-child(2n) {
  background-color: #f2f2f2; /* #f8f8f8; */
}

table tr th {
  font-weight: bold;
  border: 1px solid #cccccc;
  /* text-align: left; */
  margin: 0;
  /* padding: 6px 13px; */
  padding: 4px 6px;
}

table tr td {
  border: 1px solid #cccccc;
  /* text-align: left; */
  margin: 0;
  /* padding: 6px 13px; */
  padding: 4px 6px;
}

table tr th :first-child, table tr td :first-child {
  margin-top: 0;
}

table tr th :last-child, table tr td :last-child {
  margin-bottom: 0;
}

img {
  max-width: 100%;
}

span.frame {
  display: block;
  overflow: hidden;
}

span.frame > span {
  border: 1px solid #dddddd;
  display: block;
  float: left;
  overflow: hidden;
  margin: 13px 0 0;
  padding: 7px;
  width: auto;
}

span.frame span img {
  display: block;
  float: left;
}

span.frame span span {
  clear: both;
  color: #333333;
  display: block;
  padding: 5px 0 0;
}

span.align-center {
  display: block;
  overflow: hidden;
  clear: both;
}

span.align-center > span {
  display: block;
  overflow: hidden;
  margin: 13px auto 0;
  text-align: center;
}

span.align-center span img {
  margin: 0 auto;
  text-align: center;
}

span.align-right {
  display: block;
  overflow: hidden;
  clear: both;
}

span.align-right > span {
  display: block;
  overflow: hidden;
  margin: 13px 0 0;
  text-align: right;
}

span.align-right span img {
  margin: 0;
  text-align: right;
}

span.float-left {
  display: block;
  margin-right: 13px;
  overflow: hidden;
  float: left;
}

span.float-left span {
  margin: 13px 0 0;
}

span.float-right {
  display: block;
  margin-left: 13px;
  overflow: hidden;
  float: right;
}

span.float-right > span {
  display: block;
  overflow: hidden;
  margin: 13px auto 0;
  text-align: right;
}

code, tt {
  margin: 0 2px;
  padding: 0 5px;
  white-space: nowrap;
  border: 1px solid #eaeaea;
  background-color: #f8f8f8;
  border-radius: 3px;
}

pre code {
  margin: 0;
  padding: 0;
  white-space: pre;
  border: none;
  background: transparent;
}

.highlight pre {
  background-color: #f8f8f8;
  border: 1px solid #cccccc;
  font-size: 13px;
  line-height: 19px;
  overflow: auto;
  padding: 6px 10px;
  border-radius: 3px;
}

pre {
  background-color: #f8f8f8;
  border: 1px solid #cccccc;
  font-size: 13px;
  line-height: 19px;
  overflow: auto;
  padding: 6px 10px;
  border-radius: 3px;
}

pre code, pre tt {
  background-color: transparent;
  border: none;
}

.thumbnail {
    width: 400px;
    height: auto;
    cursor: pointer;
    }

.form-input {
    background: none;
    border: none;
    }
    
/* The Modal (background) */
.modal {
    display: none; /* Hidden by default */
    position: fixed; /* Stay in place */
    z-index: 1; /* Sit on top */
    padding-top: 100px; /* Location of the box */
    left: 0;
    top: 0;
    width: 100%; /* Full width */
    height: 100%; /* Full height */
    overflow: auto; /* Enable scroll if needed */
    background-color: rgb(0,0,0); /* Fallback color */
    background-color: rgba(0,0,0,0.4); /* Black w/ opacity */
    }
    
/* Modal Content */
.modal-content {
    background-color: #fefefe;
    margin: auto;
    padding: 20px;
    border: 1px solid #888;
    width: 80%;
    }
    
/* The Close Button */
.modal-close {
    color: #aaaaaa;
    float: right;
    font-size: 28px;
    font-weight: bold;
    }
    
.modal-close:hover,
.modal-close:focus {
    color: #000;
    text-decoration: none;
    cursor: pointer;
    }
    
.width-100{
    width: 100%;
    }
    
.bold-txt{
    font-weight: bold;
    }
    
.pointer {cursor: pointer;}

.align-left {float: left;}

.input-width {width: 9cm;}
"""

open_image_js1 = """
  <script>
function openImage(url, title) {
  var img = new Image();
  img.src = url;
  img.onload = function() {
    var win = window.open('', '_blank', 'width=' + img.width + ',height=' + img.height);
    win.document.write('<html><head><title>' + title + '</title></head><body style="overflow: scroll; margin: 0;"><img src="' + url + '"></body></html>');
  };
}
  </script>

"""

open_image_js2 = """
<script>
function openImage(url, title) {
  var img = new Image();
  img.src = url;
  img.onload = function() {
    var win = window.open('', '_blank', 'width=' + img.width + ',height=' + img.height);
    win.document.write('<html><head><title>' + title + '</title></head><body style="overflow: scroll; margin: 0;"><div style="height: ' + img.height + 'px;"></div><img src="' + url + '"></body></html>');
  };
}
</script>
"""

open_image_js3 = """
<script>
function openImage(url, title) {
  var img = new Image();
  img.src = url;
  img.onload = function() {
    var win = window.open('', '_blank', 'width=' + img.width + ',height=' + img.height);
    var doc = win.document;
    doc.open();
    doc.write('<html><head><title>' + title + '</title><style>body { margin: 0; overflow: auto; }</style></head><body><img src="' + url + '" style="width: 100%; height: auto; display: block; max-width: none;"></body></html>');
    doc.close();
  };
}
</script>
"""

# js4 this version works the best except still has about:blank
# trying url as first parameter
open_image_js = """
<script>
function openImage(url, title) {
  var img = new Image();
  img.src = url;
  img.onload = function() {
    var win = window.open('', '_blank', 'width=' + img.width + ',height=' + img.height);
    win.document.open();
    win.document.write('<html><head><title>' + title + '</title></head><body style="overflow: scroll; margin: 0;"><img src="' + url + '"></body></html>');
    win.document.close();
  };
}
</script>
"""

modal_html = """
<!-- The Modal -->
<div id="bmd-modal" class="modal">
    <!-- Modal content -->
    <div class="modal-content">
        <span class="modal-close">&times;</span>
        <p><input type="text" class="width-100" id="txt-name">
        <p><button id="btn-update">Update</button></p></p>
    </div>
</div>
"""

open_modal_js = """
<script>
    function showModal(input_id) 
    {
        // Get the modal
        var modal = document.getElementById("bmd-modal");
        
        // Prefill modal before opening
        var txt_name = document.getElementById("txt-name");
        txt_name.value = document.getElementById(input_id).value;
        
        // Open the modal
        modal.style.display = "block";
    
        // Get the <span> element that closes the modal
        var span = document.getElementsByClassName("modal-close")[0];
        
        // When the user clicks on <span> (x), close the modal
        span.onclick = function() {
            modal.style.display = "none";
        }
        
        // When the user clicks anywhere outside of the modal, close it
        window.onclick = function(event) {
            if (event.target == modal) {
                modal.style.display = "none";
            }
        }
        
        // Get the button that updates the report html
        var btn_update = document.getElementById("btn-update");
        
        // When the user clicks the update button, modify report
        btn_update.onclick = function() {
            var corrected_text = document.getElementById("txt-name").value;
            document.getElementById(input_id).value = corrected_text;
            modal.style.display = "none";
        }
    }
</script>
"""

identical_values_js = """
<script>
    var name_dict = {};
    var value_dict = {};
    var input_text_list;
    var is_first_change = true;
    window.addEventListener('load',  
    function() 
    { 
        group_identical_input_values();
    });

    function group_identical_input_values() 
    {
        input_text_list = document.querySelectorAll('input[type="text"]');
              
        // Assign value_dict. Create value to list of tag names dict.
        for (let input of input_text_list) 
        {
            if (!(input.value in value_dict))
            {
                value_dict[input.value] = [input.name];
            }
            else
            {
                value_dict[input.value].push(input.name);
            }            
        }
        
        
        for (let input of input_text_list)
        {
            
            for (let value_list of Object.values(value_dict))
                if (value_list.includes(input.name))
                {
                    name_dict[input.name] = value_list;
                    break;
                }
        }
    }
    
    function updateText(input_name) 
    {
        // Get the input
        var input = document.getElementsByName(input_name)[0];
        var new_value = input.value
        
        // Get input text set
        input_list = name_dict[input_name];
        
        // Update value
        for (let i=0; i < input_list.length; i++)
        {
            document.getElementsByName(input_list[i])[0].value = new_value;
        }
    }
    
    function toggleLabel(checkbox)
    {
        var contest = checkbox.name.replace(/_checkbox$/, "");
        var label_id = contest.concat("_label")
        
        // Get corresponding label
        var label = document.getElementById(label_id);
        
        if (checkbox.checked == true)
        {
            label.innerHTML = "Verified";
        }        
        else
        {
            label.innerHTML = "";
        }
    }
    
    function checkboxListener(checkbox)
    {        
        if (is_first_change)
        {
            alert("Save report updates by clicking the submit button at the bottom of page.");
            is_first_change = false
        }
    }
    
</script>
"""

# this one opens the image as a file.
open_image_js5 = """
<script>
function openImage(url, title) {
  var img = new Image();
  img.src = url;
  img.onload = function() {
    var win = window.open('', '_blank', 'width=' + img.width + ',height=' + img.height);
    var doc = win.document;
    doc.write('<html><head><title>' + title + '</title></head><body style="margin:0;"><iframe id="imageFrame" style="border:none;width:100%;height:100%;overflow:scroll;" allowfullscreen></iframe></body></html>');
    doc.close();
    doc.getElementById('imageFrame').src = url;
  };
}
</script>
"""


open_image_js6 = """
<script>
function openImage(url, title) {
  var img = new Image();
  img.src = url;
  img.onload = function() {
    var win = window.open(' ', '_blank', 'toolbar=0,location=0,menubar=0');
    win.document.write('<html><head><title>' + title + 
        '</title></head><body style="overflow: scroll; margin: 0;"><h1>' + title + 
        '<p><img src="' + url + 
        '" style="max-width: 100%; height: auto;"></body></html>');
    win.document.close();
  };
}
</script>
"""

old_open_image_js = """
<script>
function openImage(url) {
  var windowName = "imageWindow";
  var windowFeatures = "resizable=yes,scrollbars=yes,status=yes";
  var newWindow = window.open("", windowName, windowFeatures);
  
  // Create a new document object for the window
  var doc = newWindow.document;
  
  // Replace the default content with the full-size image
  doc.body.innerHTML = '<img src="' + url + '" style="max-width: 100%; max-height: 100%; object-fit: contain;">';
  
  // Set the window location to the URL of the full-size image
  doc.location.href = url;

  // Set the document title to the title parameter
  doc.title = title;
}
</script>
"""

